theory Code_Setup_Haskell
imports
  Util2
begin

subsection {*Code-Generator Setup*}

text {*The order of the final two arguments of @{const foldr} in
Isabelle is different from Haskell's @{text foldr}*}
definition flipped_foldr where "flipped_foldr f \<equiv> flip (foldr f)"
text {*Separate infix use and prefix use of infix operators*}
definition comp_infix where "comp_infix x y \<equiv> x \<circ> y"
definition Cons_infix where "Cons_infix x y \<equiv> x # y"

lemma foldr_flipped_foldr[code_inline]:
  "foldr f xs a = flipped_foldr f a xs"
  by (simp add: flipped_foldr_def)

lemma flip_foldr_flipped_foldr[code_inline]:
  "flip (foldr f) = flipped_foldr f"
  by (simp add: flipped_foldr_def)

lemma o_comp_infix[code_inline]: "x \<circ> y = comp_infix x y"
  by (simp add: comp_infix_def)

code_const const (Haskell "const")
code_const flip (Haskell "flip")
code_const flipped_foldr (Haskell "foldr")
code_const "op \<circ>" (Haskell "(.)")
code_const comp_infix (Haskell infixr 9 ".")
code_const "op #" (Haskell "(:)")
code_const Cons_infix (Haskell infixr 5 ":")
(*
(*Problem: Haskell's "length" is of type "[a] -> Int", whereas
  Isabelle's "length" is of type "'a list => nat" (thus having
  arbitrary precision). Hence, no direct mapping is possible/desirable.*)
code_const length (Haskell "length")
*)

end
