(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Edg
imports
  Innermost_Impl
  DP_Framework
begin

 
(* unused idea of how to implement innermost dependency graph
  , perhaps extend, perhaps throw away 

(* idea replace each variable in rule by Rule x, whereas normal
   variables x will be Orig x, and fresh variables will be Fresh i.
   in this way, the variables in rules are always disjoint *)

datatype 'v ren = Orig 'v | Rule 'v | Fresh nat

fun replace_check :: "('f,'v ren)qrules \<Rightarrow> ('f,'v ren)term \<Rightarrow> ('f,'v ren)term \<Rightarrow> bool"
where "replace_check (Q,R) ct s = (list_ex (\<lambda> (l,r). case mgu l ct of None \<Rightarrow> False | Some \<delta> \<Rightarrow> 
    NF_lhs_list Q (s \<cdot> \<delta>) \<and> list_all (\<lambda> lt. NF_lhs_list Q lt) (args (l \<cdot> \<delta>))
     ) R)"

fun icaph :: "('f,'v ren)qrules \<Rightarrow> bool \<Rightarrow> nat \<Rightarrow> ('f,'v ren)term \<Rightarrow> 'v list \<Rightarrow> ('f,'v ren)term list \<Rightarrow> ('f,'v)term \<Rightarrow> nat \<times> ('f,'v ren)term"
where "icaph _ inn n s vs _ (Var x) = (if inn \<and> x \<in> set vs then (n,Var (Orig x)) else (Suc n, Var (Fresh n)))"
  |   "icaph QR inn n s vs cargs (Fun f (t#ts)) = (let (n',ct) = icaph QR inn n s vs [] t in icaph QR inn n' s vs (ct # cargs) (Fun f ts))"
  |   "icaph QR inn n s vs cargs (Fun f []) = (let ct = Fun f (rev cargs) in (if 
            replace_check QR ct s then (Suc n, Var (Fresh n)) else (n, ct)))"

fun icap :: "('f,'v ren)qrules \<Rightarrow> bool \<Rightarrow> ('f,'v ren)term \<Rightarrow> 'v list \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v ren)term"
  where "icap QR inn s vs t = snd (icaph QR inn 0 s vs [] t)"
(* 
   usually QR is (map_vars Rule QR), 
   inn is NF_Q_R_check (True if NF(Q) \<subseteq> NF(R)), 
   s is map_vars Orig s for some DP s \<rightarrow> t,
   vs is vars_list s, and
   t is the rhs of the DP

   obtain capped term with replaces some Orig vars by Fresh vars

   TODO: use rulemaps, prove soundness
*)

fun iedg :: "('f,'v ren)qrules \<Rightarrow> bool \<Rightarrow> ('f,'v)rule \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "iedg (Q,R) inn (s,t) (u,v) = (let 
   vs = vars_term_list s;
   ss = map_vars Orig s;
   ct = icap (Q,R) inn ss vs t;
   uu = map_vars Rule u
   in (case mgu ct uu of None \<Rightarrow> False
          | Some \<delta> \<Rightarrow> NF_lhs_list Q (ss \<cdot> \<delta>) \<and> NF_lhs_list Q (uu \<cdot> \<delta>)))"

definition map_vars_trs :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'w)rules"
where "map_vars_trs f R = (map (\<lambda> (l,r). (map_vars f l, map_vars f r)) R)"

lemma iedg: assumes "NF_lhs (set Q) (s \<cdot> \<sigma>)"
  and "NF_lhs (set Q) (u \<cdot> \<delta>)"
  and "(t \<cdot> \<sigma>, u \<cdot> \<delta>) \<in> (qrstep (set Q, set R))^*"
  shows "iedg (map (map_vars Orig) Q, map_vars_trs Rule R) (NF_Q_R_check (set Q, set R)) (s,t) (u,v)"
oops
*)

no_notation Term.Hole ("\<box>")

(* If 'v is dropped we get a problem with equiv_class below. *)
datatype ('f,'v)hterm = HHole ("\<box>") | HFun 'f "('f,'v)hterm list"

declare size_simps[simp]
fun equiv_class :: "('f,'v)hterm \<Rightarrow> ('f,'v)term set" ("[_]")
  where "[\<box>] = UNIV"
      | "[HFun f hs] = {Fun f ts | ts. length ts = length hs \<and> (\<forall>i<length hs. (ts!i) \<in> [hs!i])}"
declare size_simps[simp del]

text {* An alternative induction scheme for hterms. *}
lemma
 fixes P :: "('f,'v)hterm \<Rightarrow> bool"
 assumes "P(\<box>)" and "\<And>f ss. (\<And>s. s \<in> set ss \<Longrightarrow> P s) \<Longrightarrow> P(HFun f ss)"
 shows hterm_induct[case_names HHole HFun, induct type: "hterm"]: "P t" and "\<And>t. t \<in> set ts \<Longrightarrow> P t"
using assms by (induct t and ts) auto

instantiation hterm :: ("show", type) "show" begin
  fun
    shows_hterm
  where
    "shows_hterm HHole = shows_string ''_''" |
    "shows_hterm (HFun f ts) = (
       shows f +@+ shows_list_gen id [] ''('' '', '' '')'' (map shows_hterm ts)
     )"
  definition "shows_prec (d::nat) t = shows_hterm t"
  definition "shows_list_hterm (is::('f::show,'v)hterm list) \<equiv> shows_list_aux shows is"

  lemma assoc_hterm: "shows_prec d (t::('f::show,'v)hterm) r @ s = shows_prec d t (r @ s)"
  proof (induct t arbitrary: r s)
    case HHole show ?case unfolding shows_prec_hterm_def by simp
  next
    case (HFun f ts)
    hence elt: "\<And>r s.\<forall>t\<in>set(map shows_hterm ts). id t r @ s = id t (r @ s)" unfolding shows_prec_hterm_def by auto
    show ?case
      unfolding shows_prec_hterm_def
      unfolding shows_hterm.simps assoc shows_prec_term_def o_def
      unfolding shows_list_gen_assoc[OF elt,of "[]" "''(''" "'', ''" "'')''" r s] ..
  qed
  lemma assoc_hterm_elt: "\<forall>(t::('f::show,'v)hterm)\<in>set ts. shows_prec d t r @ s = shows_prec d t (r @ s)"
    using assoc_hterm by auto
 instance by
    (intro_classes,rule assoc_hterm,unfold shows_list_hterm_def)
    (rule shows_list_aux_assoc[OF assoc_hterm_elt])
end

lemma equiv_classes_nonempty: fixes h::"('f,'v)hterm" shows "\<exists>t. t \<in> [h]"
proof (induct h)
  case HHole show ?case by simp
next
  case (HFun f hs)
  hence ex: "\<forall>i<length hs. \<exists>t. t \<in> [hs!i]" by auto
  let ?t = "\<lambda>i. (SOME t. t \<in> [hs!i])"
  let ?f = "\<lambda>s. SOME t. t \<in> [s]"
  let ?ts = "map ?f hs"
  have "\<forall>i<length hs. ?t i \<in> [hs!i]"
  proof (intro allI impI)
    fix i assume "i<length hs"
    hence "\<exists>t. t \<in> [hs!i]" using ex by simp
    thus "?t i \<in> [hs!i]" unfolding some_eq_ex[symmetric] .
  qed
  hence "\<forall>i<length hs. ?ts!i \<in> [hs!i]" using nth_map[of _ hs "?f"] by auto
  moreover have "length ?ts = length hs" by simp
  ultimately show ?case unfolding equiv_class.simps by best
qed

lemma equiv_classes_mono: assumes "[s] \<subseteq> [t]" shows 
  "[HFun f (bef @ s # aft)] \<subseteq> [HFun f (bef @ t # aft)]" (is "[HFun f ?bsa] \<subseteq> [HFun f ?bta]")
proof -
  {
    fix i u
    assume i: "i < Suc (length bef + length aft)" (is "i < ?n")
    assume elem: "u \<in> [?bsa ! i]"
    from i have "i < length bef \<or> i = length bef \<or> (\<exists> k. i = Suc (length bef) + k \<and> k < length aft)" (is "?one \<or> ?two \<or> ?three") by arith
    hence "u \<in> [?bta ! i]" 
    proof 
      assume ?one
      hence "?bta ! i = ?bsa ! i" by (auto simp: nth_append)
      thus ?thesis using elem by auto 
    next
      assume "?two \<or> ?three"
      thus ?thesis
      proof
	assume ?two
	hence "?bsa ! i = s \<and> ?bta ! i = t" by (auto simp: nth_append)
	with assms elem show ?thesis by auto
      next
	assume ?three
	from this obtain k where "i = Suc (length bef) + k \<and> k < length aft" by auto
	hence "?bta ! i = ?bsa ! i" by (auto simp: nth_append)
	with elem show ?thesis by auto	
      qed
    qed
  }
  thus ?thesis by auto
qed

definition gcMatch :: "('f,'v) hterm \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  where "gcMatch h t == (\<exists> \<sigma>. t \<cdot> \<sigma> \<in> [h])"


(* now lets head for an  executable version of gcMatch *)

fun mergeHTerms :: "('f,'v) hterm list \<Rightarrow> ('f,'v) hterm list \<Rightarrow> ('f,'v) hterm list option"
  where "mergeHTerms [] [] = Some []"
    |   "mergeHTerms [] (_#_) = None"
    |   "mergeHTerms (_#_) [] = None"
    |   "mergeHTerms (\<box> # hsl) (h # hsr) = (case mergeHTerms hsl hsr of None \<Rightarrow> None | Some res \<Rightarrow> Some (h # res))"
    |   "mergeHTerms (HFun f hs # hsl) (\<box> # hsr) = (case mergeHTerms hsl hsr of None \<Rightarrow> None | Some res \<Rightarrow> Some (HFun f hs # res))" 
    |   "mergeHTerms (HFun f fs # hsl) (HFun g gs # hsr) = 
            (if f = g then (case mergeHTerms fs gs of None \<Rightarrow> None | Some mhs \<Rightarrow> (case mergeHTerms hsl hsr of None \<Rightarrow> None | Some res \<Rightarrow> 
                             Some (HFun f mhs # res))) else None)"


  
lemma mergeHTermsComplete: "mergeHTerms hsl hsr = None \<Longrightarrow> length hsl \<noteq> length hsr \<or> (\<exists> i < length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})"
proof (induct rule: mergeHTerms.induct)
  case 1 thus ?case by simp
next
  case 2 thus ?case by simp
next
  case 3 thus ?case by simp
next
  fix hsl h hsr 
  assume rec: "mergeHTerms hsl hsr = None \<Longrightarrow> length hsl \<noteq> length hsr \<or> (\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})"
  assume res: "mergeHTerms (\<box> # hsl) (h # hsr) = None"
  show "length (\<box> # hsl) \<noteq> length (h # hsr) \<or> (\<exists>i<length (\<box> # hsl). [(\<box> # hsl) ! i] \<inter> [(h # hsr) ! i] \<subseteq> {})" 
  proof (cases "mergeHTerms hsl hsr")
    case None with rec show ?thesis by force
  next
    case (Some res)
    with res show ?thesis by simp
  qed
next
  fix f :: 'f and hs hsl hsr :: "('f,'v) hterm list"
  let ?h = "HFun f hs"
  assume rec: "mergeHTerms hsl hsr = None \<Longrightarrow> length hsl \<noteq> length hsr \<or> (\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})"
  assume res: "mergeHTerms (?h # hsl) (\<box> # hsr) = None"
  show "length (?h # hsl) \<noteq> length (\<box> # hsr) \<or> (\<exists>i<length (?h # hsl). [(?h # hsl) ! i] \<inter> [(\<box> # hsr) ! i] \<subseteq> {})" 
  proof (cases "mergeHTerms hsl hsr")
    case None with rec show ?thesis by force
  next
    case (Some res)
    with res show ?thesis by simp
  qed
next
  fix f fs hsl g gs hsr
  assume rec1: "\<lbrakk>f = g; mergeHTerms fs gs = None\<rbrakk> \<Longrightarrow> length fs \<noteq> length gs \<or> (\<exists>i<length fs. [fs ! i] \<inter> [gs ! i] \<subseteq> {})"
   and rec2: "\<And>a. \<lbrakk>f = g; mergeHTerms fs gs = Some a; mergeHTerms hsl hsr = None\<rbrakk>
            \<Longrightarrow> length hsl \<noteq> length hsr \<or> (\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})"
   and noneRes: "mergeHTerms (HFun f fs # hsl) (HFun g gs # hsr) = None"
  show "length (HFun f fs # hsl) \<noteq> length (HFun g gs # hsr) \<or>
    (\<exists>i<length (HFun f fs # hsl). [(HFun f fs # hsl) ! i] \<inter> [(HFun g gs # hsr) ! i] \<subseteq> {})"
  proof (cases "f = g")
    case False
    have "[HFun f fs] \<inter> [HFun g gs] \<subseteq> {}" 
    proof 
      fix t
      assume "t \<in> [HFun f fs] \<inter> [HFun g gs]"
      with False have False by (cases t, auto) 
      thus "t \<in> {}" ..
    qed
    hence nearly: "[(HFun f fs # hsl) ! 0] \<inter> [(HFun g gs # hsr) ! 0] \<subseteq> {}" by simp
    have "0 < length (HFun f fs # hsl)" by simp
    with nearly show ?thesis by blast 
  next
    case True
    show ?thesis 
    proof (cases "mergeHTerms fs gs")
      case None
      with `f = g` and rec1 have prop1: "length fs \<noteq> length gs \<or> (\<exists> i < length fs. [fs ! i] \<inter> [gs ! i] \<subseteq> {})" by simp
      thus ?thesis
      proof
	assume len: "length fs \<noteq> length gs" 
	have "[HFun f fs] \<inter> [HFun g gs] \<subseteq> {}"
	proof
	  fix t
	  assume "t \<in> [HFun f fs] \<inter> [HFun g gs]"
	  with len have False by (cases t, auto)
	  thus "t \<in> {}" ..
	qed
	hence nearly: "[(HFun f fs # hsl) ! 0] \<inter> [(HFun g gs # hsr) ! 0] \<subseteq> {}" by simp
	have "0 < length (HFun f fs # hsl)" by simp
	with nearly show ?thesis by blast
      next	
	assume "\<exists>i<length fs. [fs ! i] \<inter> [gs ! i] \<subseteq> {}"
	from this obtain i where len: "i < length fs" and ne: "[fs ! i] \<inter> [gs ! i] \<subseteq> {}" by force
	have "[HFun f fs] \<inter> [HFun g gs] \<subseteq> {}"
	proof
	  fix t
	  assume "t \<in> [HFun f fs] \<inter> [HFun g gs]"
	  with len and ne have False by (cases t, auto)
	  thus "t \<in> {}" ..
	qed
	hence nearly: "[(HFun f fs # hsl) ! 0] \<inter> [(HFun g gs # hsr) ! 0] \<subseteq> {}" by simp
	have "0 < length (HFun f fs # hsl)" by simp
	with nearly show ?thesis by blast
      qed
    next
      case (Some res)
      with `f = g` and rec2 and noneRes have 
	"length hsl \<noteq> length hsr \<or> (\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {})" by force
      thus ?thesis
      proof
	assume "length hsl \<noteq> length hsr" thus ?thesis by simp
      next
	assume "\<exists>i<length hsl. [hsl ! i] \<inter> [hsr ! i] \<subseteq> {}" thus ?thesis by force
      qed
    qed
  qed
qed

lemma  fixes P :: "('f,'v) hterm list \<Rightarrow> ('f,'v) hterm list \<Rightarrow> ('f,'v) hterm list  \<Rightarrow> bool"
  assumes EE: "P [] [] []" 
  and BL: "\<And>h hsl hsr mhs. \<lbrakk>P hsl hsr mhs\<rbrakk> \<Longrightarrow> P (\<box>  # hsl) (h # hsr) (h # mhs)"
  and BR: "\<And>h hsl hsr mhs. \<lbrakk>P hsl hsr mhs\<rbrakk> \<Longrightarrow> P (h  # hsl) (\<box> # hsr) (h # mhs)"
  and FF: "\<And>f ls rs hsl hsr m mhs. \<lbrakk>P ls rs m; P hsl hsr mhs\<rbrakk> \<Longrightarrow> P (HFun f ls # hsl) (HFun f rs # hsr) (HFun f m # mhs)"
  and Pre: "mergeHTerms hsl hsr = Some mhs"
  shows mergeHTerms_induct [case_names EE BL BR FF Pre]: "P hsl hsr mhs"
using Pre
proof (induct arbitrary: mhs rule: mergeHTerms.induct) 
  case 1 with EE show ?case by simp
next
  case 2 thus ?case by simp
next
  case (3 h hls mhs) thus ?case by simp
next
  case (4 hsl h hsr mhs) with BL show ?case by (cases "mergeHTerms hsl hsr", auto)
next
  case (5 f hs hsl hsr mhs) with BR show ?case by (cases "mergeHTerms hsl hsr", auto)
next
  case (6 f fs hsl g gs hsr mhs)
  from `mergeHTerms (HFun f fs # hsl) (HFun g gs # hsr) = Some mhs`
  show ?case
  proof (cases "f=g")
    case False with 6 show ?thesis by simp
  next
    case True
    {
      fix m
      assume I: "(case mergeHTerms hsl hsr of None \<Rightarrow> None | Some res \<Rightarrow> Some (HFun g m # res)) = Some mhs"
      assume II: "mergeHTerms fs gs = Some m"
      have "P (HFun g fs # hsl) (HFun g gs # hsr) mhs" using I II
      proof (cases "mergeHTerms hsl hsr", (auto)[1])
	case (Some rres)
	with I II 6 True assms show "P (HFun g fs # hsl) (HFun g gs # hsr) mhs" by auto
      qed
    }
    with 6(3) True
    show ?thesis
      by (cases "mergeHTerms fs gs", auto)
  qed
qed

lemma mergeHTermsSound: assumes "mergeHTerms hsl hsr = Some mhs" shows 
  "length hsl = length hsr \<and> length hsl = length mhs \<and> (\<forall> i < length hsl. [hsl ! i] \<inter> [hsr ! i] = [mhs ! i])"
proof (induct rule: mergeHTerms_induct)
  case EE
  thus ?case by simp
next
  case (BL h hsl hsr mhs)
  thus ?case proof (simp)
    show "\<forall> i < Suc (length hsr). [(\<box> # hsl) ! i] \<inter> [(h # hsr) ! i] = [(h # mhs) !i]"
    proof (intro allI impI)
      fix i
      assume ilen: "i < Suc (length hsr)"
      show "[(\<box> # hsl) ! i] \<inter> [(h # hsr) ! i] = [(h # mhs) !i]" using BL and ilen by (cases i, auto)
    qed
  qed
next
  case (BR h hsl hsr mhs)
  thus ?case proof (simp)
    show "\<forall> i < Suc (length hsr). [(h # hsl) ! i] \<inter> [(\<box> # hsr) ! i] = [(h # mhs) ! i]"
    proof (intro allI impI)
      fix i
      assume ilen: "i < Suc (length hsr)"
      show "[(h # hsl) ! i] \<inter> [(\<box> # hsr) ! i] = [(h # mhs) !i]" using BR and ilen by (cases i, auto)
    qed
  qed
next    
  case (FF f ls rs hsl hsr m mhs)
  thus ?case proof (simp)
    show "\<forall> i < Suc (length hsr). [(HFun f ls # hsl) ! i] \<inter> [(HFun f rs # hsr) ! i] = [(HFun f m # mhs) ! i]"
    proof (intro allI impI)
      fix i
      assume ilen: "i < Suc (length hsr)"
      show "[(HFun f ls # hsl) ! i] \<inter> [(HFun f rs # hsr) ! i] = [(HFun f m # mhs) ! i]" 
      proof (cases i)
	case (Suc j)
	with ilen and FF have "[hsl ! j] \<inter> [hsr ! j] = [mhs ! j]" by simp
	with Suc show ?thesis by simp
      next
	case 0
	from FF have lr: "length ls = length rs" and lm: "length ls = length m" and mem: "\<forall> i < length ls. [ls ! i] \<inter> [rs ! i] = [m ! i]" by auto 
	have "[HFun f ls] \<inter> [HFun f rs] = [HFun f m]" 
	proof (intro set_eqI iffI)
	  fix t 	  
	  assume "t \<in> [HFun f m]" 
	  hence "\<exists> ts. t = Fun f ts \<and> length ts = length m \<and> (\<forall> k < length m. ts ! k \<in> [m ! k])" by simp
	  from this obtain ts where "t = Fun f ts" and "length ts = length m" and "\<forall> k < length m. ts ! k \<in> [m ! k]" by force
	  with lr lm mem show "t \<in> [HFun f ls] \<inter> [HFun f rs]" by auto
	next
	  fix t 
	  assume elem: "t \<in> [HFun f ls] \<inter> [HFun f rs]"
	  hence "\<exists> tsl. t = Fun f tsl \<and> length tsl = length ls \<and> (\<forall> k < length ls. tsl ! k \<in> [ls ! k])" by simp
	  from this obtain tsl where l1: "t = Fun f tsl" and l2: "length tsl = length ls" and l3: "\<forall> k < length ls. tsl ! k \<in> [ls ! k]" by force
	  from elem have "\<exists> tsr. t = Fun f tsr \<and> length tsr = length rs \<and> (\<forall> k < length rs. tsr ! k \<in> [rs ! k])" by simp
	  from this obtain tsr where r1: "t = Fun f tsr" and r3: "\<forall> k < length rs. tsr ! k \<in> [rs ! k]" by force
	  from l1 and r1  have id: "tsl = tsr" by simp
	  have "\<forall> k < length m. tsl ! k \<in> [m ! k]"
	  proof (intro allI impI)
	    fix k 
	    assume klen: "k < length m"
	    with l3 lm have lmem: "tsl ! k \<in> [ls ! k]" by simp
	    from klen lr lm r3 id have rmem: "tsl ! k \<in> [rs ! k]" by simp
	    from lmem rmem klen mem lm show "tsl ! k \<in> [m ! k]" by force
	  qed
	  with l1 l2 lm show "t \<in> [HFun f m]" by simp
	qed
	with 0 show ?thesis by simp
      qed
    qed
  qed
next
  case Pre from assms show ?case .
qed   
	
fun merge :: "('f,'v) hterm \<Rightarrow> ('f,'v) hterm \<Rightarrow> ('f,'v) hterm option"
  where "merge hl hr = (case mergeHTerms [hl] [hr] of None \<Rightarrow> None | Some res \<Rightarrow> Some (res ! 0))"

lemma mergeComplete: assumes "merge hl hr = None" shows "[hl] \<inter> [hr] = {}"
using assms
proof (cases "mergeHTerms [hl] [hr]", auto)
  fix t
  assume n: "mergeHTerms [hl] [hr] = None" and l: "t \<in> [hl]" and r: "t \<in> [hr]"
  from n have "length [hl] \<noteq> length [hr] \<or> (\<exists> i < length [hl]. [[hl] ! i] \<inter> [[hr] ! i] \<subseteq> {})" 
    by (rule mergeHTermsComplete)
  with l r show False by force
qed

lemma mergeSound: assumes m: "merge hl hr = Some m" shows "[hl] \<inter> [hr] = [m]"
using assms
proof (cases "mergeHTerms [hl] [hr]", simp)
  fix res
  assume mt: "mergeHTerms [hl] [hr] = Some res" 
  with m have z: "m = res ! 0" by simp
  from mt have "length [hl] = length [hr] \<and> length [hl] = length res \<and> (\<forall> i < length [hl]. [[hl] ! i] \<inter> [[hr] ! i] = [res ! i])"
    by (rule mergeHTermsSound)
  with z show "[hl] \<inter> [hr] = [m]" by simp
qed
  
function gcMatchList :: "(('f,'v) hterm \<times> ('f,'v) term) list \<Rightarrow> (('f,'v) hterm \<times> 'v) list option"
  where "gcMatchList [] = Some []"
    |   "gcMatchList ((HHole, _) # ps) = gcMatchList ps"
    |   "gcMatchList ((HFun f hs,Fun g ss) # ps) = (if f = g then 
             (case decomp hs ss of Some pairs \<Rightarrow> gcMatchList (pairs @ ps) | None \<Rightarrow> None) else None)"
    |   "gcMatchList ((HFun f hs, Var x) # ps) = 
             (case gcMatchList ps of Some pairs \<Rightarrow> Some ((HFun f hs,x) # pairs) | None \<Rightarrow> None)"
by pat_completeness auto

termination 
proof -
  let ?M = "measures [\<lambda>l. list_size (size \<circ> snd) l, length]"
  show ?thesis 
  proof
    show "wf ?M" by auto    
  next
    fix hs ps 
    show "(ps, (\<box>, hs) # ps) \<in> ?M" by auto
  next
    fix f hs x ps 
    show "(ps, (HFun f hs, Var x) # ps) \<in> ?M" by auto
  next
    fix f :: 'f and hs :: "('f,'v) hterm list" and g :: 'f and ss :: "('f,'v) term list" and ps res
    assume dec: "decomp hs ss = Some res"
    show "(res @ ps, (HFun f hs, Fun g ss) # ps) \<in> ?M" 
    proof (rule measures_less, auto)
      from dec have len: "length hs = length ss" by (rule decompLength)
      from dec have res: "res = zip hs ss" by (rule decompZip)
      from len and res show "list_size (size \<circ> snd) (res @ ps) < Suc (Suc (list_size size ss + list_size (size \<circ> snd) ps))"
        by (auto simp: list_size_append list_size_map[symmetric] simp del: list_size_map)
    qed
  qed
qed

definition gcMatchL :: "(('f,'v) hterm \<times> ('f,'v) term) set \<Rightarrow> ('f,'v) subst set"
   where "gcMatchL ps == {\<sigma>. (\<forall> p \<in> ps. snd p \<cdot> \<sigma> \<in> [fst p])}"
abbreviation gcMatchLi :: "(('f,'v) hterm \<times> 'v) set \<Rightarrow> ('f,'v) subst set"
    where "gcMatchLi ps == gcMatchL ((\<lambda> p. (fst p, Var (snd p))) ` ps)"

lemma gcMatchLUnion[simp]: "gcMatchL (ps \<union> qs) = gcMatchL ps \<inter> gcMatchL qs"
  by (unfold gcMatchL_def, blast)

lemma gcMatchLInsert: "gcMatchL (insert p ps) = {\<sigma>. snd p \<cdot> \<sigma> \<in> [fst p]} \<inter> gcMatchL ps"
  by (unfold gcMatchL_def, blast)

lemma gcMatchLZip: assumes "length ts = length hs" shows 
  "{\<sigma>. \<forall>i<length ts. ts ! i \<cdot> \<sigma> \<in> [hs ! i]} = gcMatchL (set (zip hs ts))" 
using assms
proof (induct ts arbitrary:hs) 
  case Nil thus ?case by (cases hs, auto simp:gcMatchL_def)
next
  case (Cons t ss) note oCons = this
  thus ?case 
  proof (cases hs)
    case Nil with Cons show ?thesis by simp
  next
    case (Cons h hhs)
    with oCons have "length ss = length hhs" by simp
    with oCons Cons have Prop: "{\<sigma>. \<forall>i<length ss. ss ! i \<cdot> \<sigma> \<in> [hhs ! i]} = gcMatchL (set (zip hhs ss))" by simp
    have "{\<sigma>. \<forall>i<length (t # ss). (t # ss) ! i \<cdot> \<sigma> \<in> [(h # hhs) ! i]} = { \<sigma>. t \<cdot> \<sigma> \<in> [h]} \<inter> {\<sigma>. \<forall>i<length ss. (ss ! i) \<cdot> \<sigma> \<in> [hhs ! i]}"
    proof (rule set_eqI, rule iffI)
      fix s
      assume "s \<in> {\<sigma>. \<forall>i<length (t # ss). (t # ss) ! i \<cdot> \<sigma> \<in> [(h # hhs) ! i]}"
      thus "s  \<in> {\<sigma>. t \<cdot> \<sigma> \<in> [h]} \<inter> {\<sigma>. \<forall>i<length ss. ss ! i \<cdot> \<sigma> \<in> [hhs ! i]}" by force  
    next
      fix s 
      assume Prop2: "s \<in>  {\<sigma>. t \<cdot> \<sigma> \<in> [h]} \<inter> {\<sigma>. \<forall>i<length ss. ss ! i \<cdot> \<sigma> \<in> [hhs ! i]}"
      hence one: "t \<cdot> s \<in> [h]" by simp
      from Prop2 have two: "\<forall>i< length ss. ss ! i \<cdot> s \<in> [hhs ! i]" by simp
      have "\<forall> i < length (t # ss). (t # ss) ! i \<cdot> s \<in> [(h # hhs) ! i]"
      proof (rule allI, rule impI)
	fix i
	assume "i < length (t # ss)"
	with one two show "(t # ss) ! i \<cdot> s \<in> [(h # hhs) ! i]"
	  by (cases i, auto)
      qed
      thus "s \<in> {\<sigma>. \<forall>i<length (t # ss). (t # ss) ! i \<cdot> \<sigma> \<in> [(h # hhs) ! i]}" by simp
    qed
    also have "... =  { \<sigma>. t \<cdot> \<sigma> \<in> [h]} \<inter> gcMatchL (set (zip hhs ss))" using Prop by simp
    finally show ?thesis using Cons by (simp add: gcMatchLInsert)
  qed
qed

lemma gcMatchLFun[simp]: "gcMatchL (insert (HFun f hs, Fun g ts) ps) = (if f = g \<and> length hs = length ts then gcMatchL (set (zip hs ts)) \<inter> gcMatchL ps else
{})" 
proof (cases "f = g \<and> length hs = length ts")
  assume "\<not> (f = g \<and> length hs = length ts)"
  hence "f \<noteq> g \<or> length hs \<noteq> length ts" by blast
  thus ?thesis
  proof
    assume "f \<noteq> g" thus ?thesis by (unfold gcMatchL_def, simp)
  next
    assume "length hs \<noteq> length ts" thus ?thesis by (unfold gcMatchL_def, auto)
  qed
next
  assume P: "f = g \<and> length hs = length ts"
  hence "{\<sigma>. \<forall>i<length ts. ts ! i \<cdot> \<sigma> \<in> [hs ! i]} = gcMatchL (set (zip hs ts))" by (simp add: gcMatchLZip)
  with P show ?thesis by (simp add: gcMatchLInsert)
qed

    

lemma gcMatchListComplete: assumes "gcMatchList ps = None" shows "gcMatchL (set ps) = {}"
using assms 
proof (induct rule:gcMatchList.induct)
  case 1 thus ?case by simp
next
  case (2 t ps) thus ?case by (unfold gcMatchL_def, auto)
next
  case (4 f hs x ps) thus ?case by (cases "gcMatchList ps", unfold gcMatchL_def, auto)
next
  case (3 f hs g ss ps) thus ?case 
  proof (cases "f = g")
    case False thus ?thesis by (unfold gcMatchL_def, auto)
  next
    case True thus ?thesis 
    proof (cases "decomp hs ss")
      case (Some pairs)
      hence len: "length hs = length ss" by (rule decompLength)
      from Some have zip: "pairs = zip hs ss" by (rule decompZip)
      with Some `f = g` 3 have "gcMatchL (set (zip hs ss @ ps)) = {}"  by simp
      with len show ?thesis by simp
    next
      case None
      hence difflen: "length hs \<noteq> length ss" by (rule decompNone)
      thus ?thesis by simp
    qed
  qed
qed

lemma gcMatchListSound: assumes "gcMatchList ps = Some qs" shows "gcMatchL (set ps) = gcMatchLi (set qs)" 
using assms 
proof (induct arbitrary:qs rule:gcMatchList.induct)
  case 1 thus ?case by (unfold gcMatchL_def, simp)
next
  case (2 t ps) thus ?case by (unfold gcMatchL_def, auto)
next
  case (3 f hs g ss ps) thus ?case 
  proof (cases "f = g")
    case False with 3 show ?thesis by auto
  next
    case True thus ?thesis 
    proof (cases "decomp hs ss")
      case None with 3 and `f=g` show ?thesis by simp
    next
      case (Some pairs)
      hence len: "length hs = length ss" by (rule decompLength)
      from Some have zip: "pairs = zip hs ss" by (rule decompZip)
      with 3 Some True zip have "gcMatchL (set (zip hs ss @ ps)) = gcMatchLi (set qs)" by simp
      with len True show ?thesis by simp
    qed
  qed
next
  case (4 f hs x ps) thus ?case 
  proof (cases "gcMatchList ps")
    case None with 4 show ?thesis by simp
  next
    case (Some pairs) with 4 show ?thesis by (auto simp: gcMatchLInsert)
  qed
qed

fun mergeX :: "'v \<Rightarrow> ('f,'v) hterm \<Rightarrow> (('f,'v) hterm \<times> 'v) list \<Rightarrow> (('f,'v) hterm \<times> 'v) list option"
  where "mergeX x h [] = Some []"
      | "mergeX x h ((ht,y) # ps) = (if x = y then (case merge h ht of None \<Rightarrow> None | Some m \<Rightarrow> mergeX x m ps) 
                     else (case mergeX x h ps of None \<Rightarrow> None | Some pairs \<Rightarrow> Some ((ht,y) # pairs)))"

lemma mergeXComplete: assumes "mergeX x h ps = None"
  shows "gcMatchLi (insert (h,x) (set ps)) = {}"
using assms
proof (induct ps arbitrary: h)
  case Nil thus ?case by simp
next
  case (Cons p ps)
  show ?case 
  proof (cases p)
    case (Pair ht y)
    show ?thesis 
    proof (cases "x = y")
      case False
      from Cons Pair False have "mergeX x h ps = None" by (cases "mergeX x h ps", auto)
      with Cons have "gcMatchLi (insert (h, x) (set ps)) = {}" by simp
      thus ?thesis by (auto simp: gcMatchLInsert)
    next
      case True
      show ?thesis
      proof (cases "merge h ht")
	case None
	hence "[h] \<inter> [ht] = {}" by (rule mergeComplete)
	with True Pair show ?thesis by (auto simp: gcMatchLInsert)
      next
	case (Some m)
	with Pair True Cons have ind: "gcMatchLi (insert (m,x) (set ps)) = {}" by simp
	from Some have "[h] \<inter> [ht] = [m]" by (rule mergeSound)
	with Pair ind True show ?thesis by (auto simp:gcMatchLInsert)
      qed
    qed
  qed
qed

lemma mergeXSound: assumes "mergeX x h ps = Some rps"
  shows "\<exists> rh. (gcMatchLi (insert (h,x) (set ps)) = gcMatchLi (insert (rh,x) (set rps)) \<and> x \<notin> snd ` set rps)"
using assms
proof (induct ps arbitrary: h rps)
  case Nil thus ?case by auto
next
  case (Cons p ps)
  show ?case 
  proof (cases p)
    case (Pair ht y)
    show ?thesis 
    proof (cases "x = y")
      case False
      show ?thesis 
      proof (cases "mergeX x h ps")
	case None with Cons Pair False show ?thesis by simp
      next
	case (Some pairs)
	with Cons Pair False Some 
	obtain m where "gcMatchLi (insert (h,x) (set ps)) = gcMatchLi (insert (m,x) (set pairs)) \<and> x \<notin> snd ` set pairs" 
          and "rps = (ht,y) # pairs" by force 
	with False Pair show ?thesis by (auto simp: gcMatchLInsert)
      qed
    next
      case True 
      show ?thesis
      proof (cases "merge h ht")
	case None with Pair Cons True show ?thesis by simp
      next
	case (Some m)
	with Pair True Cons obtain rh where 
	  ind: "gcMatchLi (insert (m,x) (set ps)) = gcMatchLi (insert (rh,x) (set rps)) \<and> x \<notin> snd ` set rps" by force
	from Some have "[h] \<inter> [ht] = [m]" by (rule mergeSound)
	with Pair ind True show ?thesis by (auto simp:gcMatchLInsert)
      qed
    qed
  qed
qed

lemma mergeXSize[simp]: assumes "mergeX x h ps = Some rps" shows "length ps \<ge> length rps"
using assms
proof (induct ps arbitrary: h rps)
  case Nil thus ?case by simp
next
  case (Cons p ps)
  show ?case
  proof (cases p)
    case (Pair ht y)
    show ?thesis
    proof (cases "x = y")
      case False
      with Cons Pair show ?thesis by (cases "mergeX x h ps", auto)
    next
      case True
      with Cons Pair show ?thesis by (cases "merge h ht", auto, force)
    qed
  qed
qed  

fun mergeAll :: "(('f,'v) hterm \<times> 'v) list \<Rightarrow> bool"
  where "mergeAll [] = True"
     |  "mergeAll ((h,x) # ps) = (case mergeX x h ps of None \<Rightarrow> False | Some mps \<Rightarrow> mergeAll mps)"

lemma mergeAllComplete: fixes ps :: "(('f,'v) hterm \<times> 'v) list"
  shows "\<not> mergeAll ps \<Longrightarrow> gcMatchLi (set ps) = {}"
proof (induct rule: length_induct)
  fix pps :: "(('f,'v) hterm \<times> 'v) list"
  assume ind: "\<forall> mps. length (mps :: (('f,'v) hterm \<times> 'v) list) < length pps \<longrightarrow> \<not> mergeAll mps \<longrightarrow> gcMatchLi (set mps) = {}" and
    res: "\<not> mergeAll pps"
  show "gcMatchLi (set pps) = {}"
  proof (cases pps)
    case Nil with res show ?thesis by simp
  next
    case (Cons p ps)
    show ?thesis
    proof (cases p)
      case (Pair h x)
      show ?thesis
      proof (cases "mergeX x h ps")
	case None
	hence "gcMatchLi (insert (h,x) (set ps)) = {}" by (rule mergeXComplete)
	with Cons Pair show ?thesis by simp
      next
	case (Some mps)
	hence "\<exists> rh. (gcMatchLi (insert (h,x) (set ps)) = gcMatchLi (insert (rh,x) (set mps)) \<and> x \<notin> snd ` set mps)" by (rule mergeXSound)
	from this obtain rh where m: "gcMatchLi (insert (h,x) (set ps)) = gcMatchLi (insert (rh,x) (set mps))" by auto
	from Some have "length mps \<le> length ps" by simp
	with Cons have len: "length mps < length pps" by simp
	from Cons Pair Some res have "\<not> mergeAll mps" by auto
	with len ind have "gcMatchLi (set mps) = {}" by auto
	with m Pair Cons show ?thesis by (simp add: gcMatchLInsert)
      qed
    qed
  qed
qed

lemma mergeAllSound: fixes ps :: "(('f,'v) hterm \<times> 'v) list"
  shows "mergeAll ps \<Longrightarrow> \<exists> \<sigma>. \<sigma> \<in> gcMatchLi (set ps)"
proof (induct rule: length_induct)
  fix pps :: "(('f,'v) hterm \<times> 'v) list"
  assume ind: "\<forall> mps. length (mps :: (('f,'v) hterm \<times> 'v) list) < length pps \<longrightarrow> mergeAll mps \<longrightarrow> (\<exists> \<sigma>. \<sigma> \<in> gcMatchLi (set mps))" and
    res: "mergeAll pps"
  show "\<exists> \<sigma> . \<sigma> \<in> gcMatchLi (set pps)"
  proof (cases pps)
    case Nil with res show ?thesis by (unfold gcMatchL_def, simp)
  next
    case (Cons p ps)
    show ?thesis
    proof (cases p)
      case (Pair h x)
      show ?thesis
      proof (cases "mergeX x h ps")
	case None with Cons Pair res show ?thesis by simp
      next
	case (Some mps)
	hence "\<exists> rh. (gcMatchLi (insert (h,x) (set ps)) = gcMatchLi (insert (rh,x) (set mps)) \<and> x \<notin> snd ` set mps)" by (rule mergeXSound)
	from this obtain rh where m: "gcMatchLi (insert (h,x) (set ps)) = gcMatchLi (insert (rh,x) (set mps))" 
	  and noX: "x \<notin> snd ` set mps" by auto
	from Some have "length mps \<le> length ps" by simp
	with Cons have len: "length mps < length pps" by simp
	from Cons Pair Some res have "mergeAll mps" by auto
	with len ind obtain \<sigma> where sigma: "\<sigma> \<in> gcMatchLi (set mps)" by auto
	show ?thesis 
	proof -
	  have "\<exists> t. t \<in> [rh]" by (rule equiv_classes_nonempty)
	  from this obtain t where trh: "t \<in> [rh]" ..
	  let ?delta = "(\<lambda>y. if x = y then t else \<sigma> y)"
	  have "?delta \<in> gcMatchLi (insert (rh, x) (set mps))" using trh 
	  proof (simp add: gcMatchLInsert)
	    show "?delta \<in> gcMatchLi (set mps)" using sigma noX by (auto simp: gcMatchL_def)
	  qed
	  with m Pair Cons show ?thesis by auto
	qed
      qed
    qed
  qed
qed

fun gcMatchI :: "('f,'v) hterm \<Rightarrow> ('f,'v) term \<Rightarrow> bool"
  where "gcMatchI h t = (case gcMatchList [(h,t)] of None \<Rightarrow> False | Some pairs \<Rightarrow> mergeAll pairs)"

lemma gcMatchISound[simp]: "gcMatchI = gcMatch"
proof (intro ext)
  fix h t
  have "gcMatch h t = gcMatchI h t"
  proof (cases "gcMatchList [(h,t)]")
    case None
    hence "gcMatchL (set [(h,t)]) = {}" by (rule gcMatchListComplete)
    with None show ?thesis by (simp add: gcMatchL_def gcMatch_def)
  next
    case (Some pairs)
    hence one: "gcMatchL (set [(h,t)]) = gcMatchLi (set pairs)" by (rule gcMatchListSound)
    show ?thesis
    proof (cases "mergeAll pairs")
      case False
      hence "gcMatchLi (set pairs) = {}" by (rule mergeAllComplete)
      with one Some have "gcMatchL {(h,t)} = {}" by simp
      hence "\<not> gcMatch h t" by (simp add: gcMatchL_def gcMatch_def)
      with Some one False show ?thesis by simp
    next
      case True
      hence "\<exists> \<sigma>. \<sigma> \<in> gcMatchLi (set pairs)" by (rule mergeAllSound)
      with one Some have "\<exists> \<sigma> . \<sigma> \<in> gcMatchL {(h,t)}" by auto
      hence "gcMatch h t" by (simp add: gcMatchL_def gcMatch_def)
      with Some one True show ?thesis by simp
    qed
  qed
  thus "gcMatchI h t = gcMatch h t" by simp
qed

declare gcMatchI.simps[simp del]
    

  

(* a semantic version of cap which is not executable *)

fun cap :: "('f,'v)trs \<Rightarrow> ('f,'v) term  \<Rightarrow> ('f,'v) hterm"
  where "cap _ (Var _) = \<box>"
    |   "cap R (Fun f ts) = 
            (let h = HFun f (map (cap R) ts) in (
                if (\<exists> r \<in> R. gcMatch h (fst r)) then \<box> else h
            ))"

(* an executable version of cap *)
fun capI :: "('f,'v)rules \<Rightarrow> ('f,'v) term  \<Rightarrow> ('f,'v) hterm"
  where "capI _ (Var _) = \<box>"
    |   "capI R (Fun f ts) = 
            (let h = HFun f (map (capI R) ts) in (
                if list_ex (\<lambda> r. gcMatchI h (fst r)) R then \<box> else h
            ))"

(* and a version for the reversed TRS *)
fun capR :: "('f,'v)rules \<Rightarrow> ('f,'v) term  \<Rightarrow> ('f,'v) hterm"
  where "capR _ (Var _) = \<box>"
    |   "capR R (Fun f ts) = 
            (let h = HFun f (map (capR R) ts) in (
                if list_ex (\<lambda> r. gcMatchI h (snd r)) R  then \<box> else h
            ))"


(* and a version with rule maps *)
fun capRM2 :: "('f \<times> nat \<Rightarrow> ('f,'v)rules) \<Rightarrow> ('f,'v) term \<Rightarrow> ('f,'v) hterm"
where "capRM2 _ (Var _) = \<box>"
  |   "capRM2 rm (Fun f ts) = 
            (let h = HFun f (map (capRM2 rm) ts); n = length ts in (
                if list_ex (\<lambda> r. gcMatchI h (fst r)) (rm (f,n)) then \<box> else h
            ))"

fun capRM :: "('f,'v)rule_map \<Rightarrow> ('f,'v) term  \<Rightarrow> ('f,'v) hterm"
where "capRM R = (case R of None \<Rightarrow> \<lambda> t. \<box> | Some rm \<Rightarrow> capRM2 rm)"

lemma capRM2Sound: fixes t :: "('f,'v)term"
  assumes rulemap: "computeRuleMap R = Some rm"
  shows "capRM2 rm t = cap (set R) t"
proof -
  from rulemap have noLeftVars: "\<forall> (l, r) \<in> set R. root l \<noteq> None" using computeRuleMapSound2 by force
  from rulemap have rules: "\<And> f n. set (rm (f,n)) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f \<and> num_args l = n}" using computeRuleMapSound by force
  show "capRM2 rm t = cap (set R) t" 
  proof (induct t, simp)
    case (Fun f ts)
    let ?hArgs = "(map (cap (set R)) ts) :: ('f,'v) hterm list"
    let ?hIArgs = "(map (capRM2 rm) ts) :: ('f,'v) hterm list "
    let ?h = "HFun f ?hArgs"
    let ?hI = "HFun f ?hIArgs"
    from Fun have heq: "?hI = ?h" by simp
    hence hArgsEq: "?hIArgs = ?hArgs" by simp
    let ?EQ = "\<lambda>(h1 :: ('f,'v) hterm) (h2 :: ('f,'v) hterm) n . ((\<exists> r \<in> set R. gcMatch h1 (fst r)) = (list_ex (\<lambda> r. gcMatchI h2 (fst r)) (rm (f,n)) ))"
    have "!! h1 h2 . h1 = h2 \<Longrightarrow> ?EQ (HFun f h1) (HFun f h2) (length h2)"
    proof -
      fix h1 h2 :: "('f,'v)hterm list" 
      assume h12: "h1 = h2"
      let ?mat = "\<lambda> r . gcMatchI (HFun f h2) (fst r)"
      have "(list_ex (\<lambda> r. ?mat r) (rm (f,length h2))) = (\<exists> r \<in> set (rm (f,length h2)). ?mat r)" by (simp add: list_ex_iff)
      also have "\<dots> = (\<exists> r \<in> set R. ?mat r)"
      proof 
	assume "\<exists> r \<in> set (rm (f,length h2)). ?mat r"
	from this obtain r where "r \<in> set (rm (f,length h2)) \<and> ?mat r" by auto
	with rules have "r \<in> set R \<and> ?mat r" by auto
	thus "\<exists> r \<in> set R. ?mat r" by auto
      next
	assume "\<exists> r \<in> set R. ?mat r"
	from this obtain r where A: "r \<in> set R \<and> ?mat r" by auto
	with noLeftVars obtain g ts' rhs where B: "r = (Fun g ts', rhs)" by (cases r, cases "fst r", auto)
	from A B have "f = g \<and> length ts' = length h2" by (cases "f=g", auto simp: gcMatch_def)
	with A B rules have "(Fun f ts', rhs) \<in> set (rm (f,length h2)) \<and> ?mat (Fun f ts', rhs)" by simp        
	thus "\<exists> r \<in> set (rm (f,length h2)). ?mat r" by blast
      qed
      also have "\<dots> = (\<exists> r \<in> set R. gcMatch (HFun f h1) (fst r))" using  h12 by auto
      finally show "?EQ (HFun f h1) (HFun f h2) (length h2)" by simp
    qed
    from this[OF hArgsEq[symmetric]] have a: "?EQ ?h ?hI (length ts)" by auto
    show ?case using if_weak_cong[OF a, where x = "HHole" and y = "HFun f (map (cap (set R)) ts)",symmetric] unfolding capRM2.simps cap.simps Let_def heq gcMatchISound by (force)
  qed
qed
    

lemma capRMSound:   fixes R :: "('f,'v) rule list"
  and t :: "('f,'v) term"
  shows "capRM (computeRuleMap R) t = cap (set R) t"
proof (cases "computeRuleMap R")
  case None
  hence left: "capRM (computeRuleMap R) t = \<box>" by simp
  from None have "\<exists> (l, r) \<in> set R. root l = None" using computeRuleMapSound2 by auto
  from this obtain l r where "(l,r) \<in> set R \<and> root l = None" by auto
  from this obtain x where rule: "(Var x,r) \<in> set R" by (cases l, auto)
  have "cap (set R) t = \<box>" 
  proof (cases t, simp)
    case (Fun f ts)
    have "gcMatch (HFun f (map (cap (set R)) ts)) (fst (Var x, r))" by (simp only: gcMatchISound[symmetric] gcMatchI.simps, simp)
    with rule Fun show ?thesis by (auto simp: Let_def, blast)
  qed
  with left show ?thesis by simp
next
  case (Some rm)
  with capRM2Sound show ?thesis by auto
qed  
declare capRM.simps[simp del]

lemma capISound:   fixes R :: "('f,'v) rule list"
  and t :: "('f,'v) term"
  shows "capI R t = cap (set R) t"
proof (induct t)
    case (Var x) thus ?case by simp
  next
    case (Fun f ts)
    let ?h = "HFun f (map (cap (set R)) ts) :: ('f,'v) hterm"
    let ?hI = "HFun f (map (capI R) ts) :: ('f,'v) hterm"
    from Fun have heq: "?hI = ?h" by simp
    let ?EQ = "\<lambda>(h1 :: ('f,'v) hterm) (h2 :: ('f,'v) hterm) . ((\<exists> r \<in> set R. gcMatch h1 (fst r)) = (list_ex (\<lambda> r. gcMatchI h2 (fst r)) R ))"
    have "!! h1 h2 . h1 = h2 \<Longrightarrow> ?EQ h1 h2 "
    proof -
      fix h1 h2 :: "('f,'v)hterm" 
      show "h1 = h2 \<Longrightarrow> ?EQ h1 h2" by (induct R, auto)
    qed
    hence "?h = ?hI \<Longrightarrow> ?EQ ?h ?hI" .
    with heq have a: "?EQ ?h ?hI" by simp
    show ?case using if_weak_cong[OF a, where x = "HHole" and y = "HFun f (map (cap (set R)) ts)",symmetric] unfolding capI.simps cap.simps Let_def heq gcMatchISound by force
qed

lemma nrrstep_seq_imp_arg_rstep_seq: 
  assumes "(s,t) \<in> (nrrstep R)^*"
  shows "root s = root t \<and> num_args s = num_args t \<and> (\<forall> i < num_args s. (s |_ <i>, t |_ <i> ) \<in> (rstep R)^*)"
using assms 
proof (induct)
  case base show ?case by simp
next
  case (step u t)
  hence root: "root s = root u" and num: "num_args s = num_args u" by auto
  from `(u,t) \<in> nrrstep R` have fu: "\<exists> f us ts. u = Fun f us \<and> t = Fun f ts" by (rule nrrstep_equiv_root)
  with root have eqroot: "root s = root t" by auto
  from step have arg: "\<forall> i < num_args s. (s |_ <i> , u |_ <i>) \<in> (rstep R)^*" by simp
  from `(u,t) \<in> nrrstep R` have 
    "\<exists>j<num_args u. num_args u = num_args t \<and> (u |_ <j>, t |_ <j>) \<in> rstep R \<and> (\<forall>k. j \<noteq> k \<longrightarrow> u |_ <k> = t |_ <k>)"
    by (rule nrrstep_imp_subt_rstep)
  from this obtain j where jnum: "j < num_args u" and num2: "num_args u = num_args t" 
    and utj: "(u |_ <j>, t |_ <j>) \<in> rstep R"
    and utk: "\<forall>k. j \<noteq> k \<longrightarrow> u |_ <k> = t |_ <k>" by force
  from num and num2 have num3: "num_args s = num_args t" by simp
  have "\<forall> i < num_args s. (s |_ <i> , t |_ <i>) \<in> (rstep R)^*" 
  proof (intro allI impI) 
    fix i
    assume inum: "i < num_args s"
    show "(s |_ <i> , t|_<i>) \<in> (rstep R)^*" 
    proof (cases "j = i")
      case False
      with utk have last: "u |_ <i> = t |_ <i>" by simp
      from inum arg have "(s|_<i>, u|_<i>) \<in> (rstep R)^*" by simp
      from this and last show ?thesis by simp 
   next
      case True
      with utj have last: "(u|_<i>,t|_<i>) \<in> rstep R" by simp
      from inum arg have "(s|_<i>, u|_<i>) \<in> (rstep R)^*" by simp
      from this and last show ?thesis by simp
    qed
  qed
  from this and num3 and eqroot show ?case by simp
qed


lemma capSound: assumes "(s \<cdot> \<sigma>,t) \<in> (rstep R)^*" shows 
  "t \<in> [cap R s]"
using assms
proof (induct s arbitrary: t rule: term_induct)
  case (Var x) thus ?case by simp
next
  case (Fun f ss)
  let ?h = "HFun f (map (cap R) ss)" 
  show ?case 
  proof (cases "cap R (Fun f ss)")
    case HHole
    thus ?thesis by simp
  next
    case (HFun g hs)
    hence res: "cap R (Fun f ss) = ?h"
      by (cases "\<exists> r \<in> R. gcMatch ?h (fst r)", auto simp: Let_def)
    have nrr: "!! u. (Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^* \<Longrightarrow> u \<in> [?h]" 
    proof -
      fix u
      assume "(Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^*"
      hence cond: "root (Fun f ss \<cdot> \<sigma>) = root u \<and> num_args (Fun f ss \<cdot> \<sigma>) = num_args u \<and> (\<forall> i < num_args (Fun f ss \<cdot> \<sigma>). (Fun f ss \<cdot> \<sigma> |_ <i>, u |_ <i> ) \<in> (rstep R)^*)"	
	by (rule nrrstep_seq_imp_arg_rstep_seq)
      hence "root (Fun f ss) = root u" by simp
      hence "\<exists> us. u = Fun f us" by (cases u, auto)
      from this obtain us where u: "u = Fun f us" by force
      with cond have len: "length ss = length us" by simp      
      {
	fix i
	assume ilen: "i < length us"
	with len have "<i> \<in> poss (Fun f ss)" by simp
	hence "Fun f ss \<cdot> \<sigma>|_<i> = Fun f ss |_<i> \<cdot> \<sigma>" by simp
	with ilen len cond have one: "(Fun f ss |_<i> \<cdot> \<sigma>, u|_<i>) \<in> (rstep R)^*" by simp	
	from ilen len have "Fun f ss |_<i> \<in> set ss" by simp
	with one Fun have "u|_<i> \<in> [cap R (Fun f ss |_<i>)]" by simp
	with u have "us ! i \<in> [cap R (ss ! i)]" by simp
      }
      with len have "Fun f us \<in> [?h]" 
        by (auto)
      with u show "u \<in> [?h]" by simp
    qed    
    have "rstep R = rrstep R \<union> nrrstep R" by (rule rstep_iff_rrstep_or_nrrstep)
    from this and `(Fun f ss \<cdot> \<sigma>, t) \<in> (rstep R)^*`
    have  "(Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^* \<or> (Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^* O (rrstep R) O (rstep R)^*" by (rule firstStep)
    thus ?thesis
    proof 
      assume "(Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^*"
      with nrr res show ?thesis by (auto)
    next
      assume "(Fun f ss \<cdot> \<sigma>, t) \<in> (nrrstep R)^* O rrstep R O (rstep R)^*"
      from this obtain u v where fu: "(Fun f ss \<cdot> \<sigma>, u) \<in> (nrrstep R)^*" and uv: "(u,v) \<in> rrstep R" by force
      from fu nrr have uh: "u \<in> [?h]" by simp
      from uv have "\<exists> l r \<sigma>. (l,r) \<in> R \<and> l \<cdot> \<sigma> = u \<and> r \<cdot> \<sigma> = v" by (rule rrstep_imp_rule_subst)
      from this obtain l r \<sigma> where lrR: "(l, r) \<in> R" and gcMatch: "l \<cdot> \<sigma> = u" by force
      from uh and gcMatch  have "gcMatch ?h l" by (unfold gcMatch_def, auto)
      with lrR have "cap R (Fun f ss) = \<box>" by (simp add: Let_def, force)
      thus ?thesis by simp
    qed
  qed
qed

lemma capInstanceSubset: "[cap R (t \<cdot> \<sigma>)] \<subseteq> [cap R t]" 
proof (induct t, simp)
  case (Fun f ts)
  let ?h = "HFun f (map (cap R) ts)"
  let ?hs = "HFun f (map (cap R) (map (\<lambda> t. (t \<cdot> \<sigma>)) ts))"
  let ?hss = "HFun f (map (\<lambda> t. cap R (t \<cdot> \<sigma>)) ts)"
  show ?case 
  proof (cases "\<exists> r \<in> R. gcMatch ?h (fst r)", simp)
    case False
    hence nMatch: "\<forall> r \<in> R. \<not> gcMatch ?h (fst r)" by auto
    {
      fix l
      assume "gcMatch ?hs l"
      from this obtain \<delta> where "l \<cdot> \<delta> \<in> [?hs]" unfolding gcMatch_def by auto
      hence "l \<cdot> \<delta> \<in> [?h]" using Fun unfolding gcMatch_def by force
      hence "gcMatch ?h l" unfolding gcMatch_def by auto
    }
    with nMatch have "\<forall> r \<in> R. \<not> gcMatch ?hs (fst r)" by auto
    with `\<not> (\<exists> r \<in> R. gcMatch ?h (fst r))`
    have id: "cap R (Fun f ts) = ?h \<and> cap R (Fun f ts \<cdot> \<sigma>) = ?hss" by auto
    from Fun have "[?hss] \<subseteq> [?h]" by (force simp: nth_mem)
    with id show ?thesis by simp
  qed
qed

lemma cap_refl: "t \<in> [cap R t]"
  by  (rule capSound[of _ Var], simp)

lemma cap_lhs: assumes "(l,r) \<in> R"
  shows "cap R (l \<cdot> \<sigma>) = \<box>"
proof (cases l)
  case (Fun f ts)
  show ?thesis
  proof (unfold Fun, simp add: Let_def o_def, intro bexI[OF _ assms], simp add: Fun)
    show "gcMatch (HFun f (map (\<lambda> x. cap R (x \<cdot> \<sigma>)) ts)) (Fun f ts)"
      by  (unfold gcMatch_def, rule exI[of _ \<sigma>], unfold equiv_class.simps, simp add: cap_refl)
  qed
next
  case (Var x)
  {
    fix t
    have "cap R t = \<box>"
    proof (induct t)
      case (Var y)
      thus ?case by auto
    next
      case (Fun f ts)
      show ?case
        unfolding cap.simps Let_def
        by (clarsimp, rule bexI[OF _ assms], unfold Var, simp add: gcMatch_def,
        rule exI[of _ "(\<lambda>_. Fun f ts)"], simp add: cap_refl)
    qed  
  }
  thus ?thesis by simp
qed

abbreviation capBelow :: "('f,'v)rules \<Rightarrow> 'f \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)hterm"
where "capBelow R f ts \<equiv> HFun f (map (cap (set R)) ts)"

fun matchCapBelow :: "('f,'v)term \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "matchCapBelow l R (Fun f ts) = ( gcMatch (capBelow R f ts) l)" 

lemma gcMatchBelow: assumes match: "gcMatch (HFun f (bef @ s # aft)) l" (is "gcMatch ?s l")
  and subset: "[s] \<subseteq> [t]"
  shows "gcMatch (HFun f (bef @ t # aft)) l" (is "gcMatch ?t l")
proof -
  from subset have "[?s] \<subseteq> [?t]" by (rule equiv_classes_mono)
  with match show ?thesis unfolding gcMatch_def by blast
qed


lemma cap_rewrite: fixes t :: "('f,'v)term"
  assumes "(t, s) \<in> rstep R"
  shows "[cap R s] \<subseteq> [cap R t]"
using assms
proof (induct)
  fix C \<sigma> l r assume "(l, r) \<in> R"
  thus "[cap R C\<langle>r \<cdot> \<sigma>\<rangle>] \<subseteq> [cap R C\<langle>l \<cdot> \<sigma>\<rangle>]"
  proof (induct C)
    case (More f bef D aft)
    hence ind: "[cap R D\<langle>r \<cdot> \<sigma>\<rangle>] \<subseteq> [cap R D\<langle>l \<cdot> \<sigma>\<rangle>]" by auto
    let ?h = "\<lambda>t. HFun f (map (cap R) bef @ cap R D\<langle>t\<rangle> # map (cap R) aft)"
    from ind have res: "\<And>t. gcMatch (?h (r \<cdot> \<sigma>)) t \<Longrightarrow> gcMatch (?h (l \<cdot> \<sigma>)) t"
      using gcMatchBelow by force
    show ?case
    proof (cases "\<exists>lr\<in>R. gcMatch (?h (l \<cdot> \<sigma>)) (fst lr)")
      case False 
      hence one: "\<not> (\<exists> lr \<in> R. gcMatch (?h (r \<cdot> \<sigma>)) (fst lr))" using res by blast
      from ind have subset: "[?h (r \<cdot> \<sigma>)] \<subseteq> [?h (l \<cdot> \<sigma>)]" by (rule equiv_classes_mono)
      with one show ?thesis by (auto simp: Let_def)
    next
      case True thus ?thesis by simp
    qed
  next
    case Hole
    have "cap R (l \<cdot> \<sigma>) = \<box>"
    proof (cases l)
      case (Var x)
      {
        fix h
	from equiv_classes_nonempty obtain s :: "('f,'v)term" where "s \<in> [h]" by auto
	with Var have "l \<cdot> (subst x s) \<in> [h]" unfolding subst_def by auto
	hence "gcMatch h (fst (l,r))" unfolding gcMatch_def by (simp, blast)
      }
      with `(l, r) \<in> R` show ?thesis by (cases "l \<cdot> \<sigma>", auto simp: Let_def, blast)
    next
      case (Fun f ll)
      hence one: "l \<cdot> \<sigma> = Fun f (map (\<lambda> u. u \<cdot> \<sigma>) ll)" (is "_ = Fun f ?ls") by auto
      {
        fix i 
        have "ll ! i \<cdot> \<sigma> \<in> [cap R (ll ! i \<cdot> \<sigma>)]"
          using capSound[OF rtrancl_refl, of "ll!i \<cdot> \<sigma>" "Var"] by simp
      }
      with one have "(fst (l,r)) \<cdot> \<sigma> \<in> [HFun f (map (\<lambda>x. cap R (x \<cdot> \<sigma>)) ll)]" by auto
      with one and `(l, r) \<in> R`
        show ?thesis by (auto simp: Let_def o_def, unfold gcMatch_def, blast)
    qed
    thus "[cap R Hole\<langle>r \<cdot> \<sigma>\<rangle>] \<subseteq> [cap R Hole\<langle>l \<cdot> \<sigma>\<rangle>]" by simp
  qed
qed


lemma cap_rewrites:
  assumes  "(s, t) \<in> (rstep R)^*" 
  shows "[cap R t] \<subseteq> [cap R s]"
using assms
proof (induct)
  case base thus ?case by simp
next
  case (step t u)
  from step(3) cap_rewrite[OF step(2)] show ?case by auto
qed

lemma gcMatchCapSound: 
  assumes derivation: "(t \<cdot> \<delta>, s \<cdot> \<sigma>) \<in> (rstep R)^*"
  shows "gcMatch (cap R t) s"
proof -
  from derivation have "s \<cdot> \<sigma> \<in> [cap R t]" by (rule capSound)
  thus ?thesis by (unfold gcMatch_def, auto)
qed

lemma gcMatchCapRMSound: 
  assumes derivation:  "(t \<cdot> \<delta>, s \<cdot> \<sigma>) \<in> (rstep (set R))^*"
  shows "gcMatch (capRM (computeRuleMap R) t) s"
proof -
  from derivation have "gcMatch (cap (set R) t) s" by (rule gcMatchCapSound)
  thus "gcMatch (capRM (computeRuleMap R) t) s" by (simp add: capRMSound)
qed


lemma gcMatchCapRRMSound: 
  assumes derivation:  "(t \<cdot> \<sigma>, s \<cdot> \<delta>) \<in> (rstep (set R))^*"
  shows "gcMatch (capRM (computeRuleMap (reverse_trs R)) s) t"
proof -
  let ?R1 = "reverse_trs R"
  have rev: "set ?R1 = (set R)^-1" unfolding reverse_trs_def by (induct R, auto)
  have id: "((rstep (set R))^-1)^* = ((rstep (set R))^*)^-1" by (rule rtrancl_converse)
  from derivation have "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> ((rstep (set R))^*)^-1" by simp
  hence "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> (rstep ((set R)^-1))^*" using id by simp
  with rev have "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> (rstep (set ?R1))^*" by simp
  hence mapResult: "gcMatch (capRM (computeRuleMap ?R1) s) t" by (rule gcMatchCapRMSound)
  thus ?thesis by simp
qed

lemma gcMatchCapISound: 
  assumes derivation:  "(t \<cdot> \<delta>, s \<cdot> \<sigma>) \<in> (rstep (set R))^*"
  shows "gcMatch (capI R t) s"
proof -
  from derivation have "gcMatch (cap (set R) t) s" by (rule gcMatchCapSound)
  thus "gcMatch (capI R t) s" by (simp add: capISound)
qed
  
lemma gcMatchCapRSound :
  assumes derivation: "(t \<cdot> \<sigma>, s \<cdot> \<delta>) \<in> (rstep (set R))^*"
  shows "gcMatch (capR R s) t"
proof -
  let ?R1 = "map (\<lambda> lr. (snd lr, fst lr)) R"
  have rev: "set ?R1 = (set R)^-1" by (induct R, auto)
  have id: "((rstep (set R))^-1)^* = ((rstep (set R))^*)^-1" by (rule rtrancl_converse)
  from derivation have "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> ((rstep (set R))^*)^-1" by simp
  hence "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> (rstep ((set R)^-1))^*" using id by simp
  with rev have "(s \<cdot> \<delta>, t \<cdot> \<sigma>) \<in> (rstep (set ?R1))^*" by simp
  hence mapResult: "gcMatch (capI ?R1 s) t" by (rule gcMatchCapISound)
  have "capR R s = capI ?R1 s"
  proof (induct s)
    case (Var x) thus ?case by simp
  next
    case (Fun f ss) 
    hence eq1: "map (capR R) ss = map (capI ?R1) ss" by simp
    have eq2: "\<And>h. list_ex (\<lambda> r. gcMatch h (snd r)) R  = list_ex (\<lambda> r. gcMatch h (fst r)) ?R1"
    proof -
      fix h
      show "list_ex (\<lambda> r. gcMatch h (snd r)) R = list_ex (\<lambda> r. gcMatch h (fst r)) ?R1"
	by (induct R, auto)
    qed
    have "capR R (Fun f ss) =  (let h = HFun f (map (capI ?R1) ss) in 
                (if list_ex (\<lambda> r. gcMatch h (snd r) ) R then \<box> else h))" by (simp add: eq1)
    also have "... = capI ?R1 (Fun f ss)" using eq2 by (simp add: Let_def) 
    finally show ?case .
  qed
  with mapResult show ?thesis by simp
qed

(* now that cap has been defined we come to the definition of the 
   estimated dependency graph *)

fun checkEdg :: "('f,'v) rule list \<Rightarrow> (('f,'v)rule \<times> ('f,'v)rule) \<Rightarrow> bool"
  where "checkEdg R ((s,t),(u,v)) = (gcMatchI (capI R t) u   \<and>    gcMatchI (capR R u) t)"


lemma checkEdgSound[simp]: assumes seq: "(t \<cdot> \<sigma>, u \<cdot> \<delta>) \<in> (rstep (set R))^*" shows 
  "checkEdg R ((s,t),(u,v))"
proof -
  from seq have forward: "gcMatch (capI R t) u" by (rule gcMatchCapISound)
  from seq have backward: "gcMatch (capR R u) t" by (rule gcMatchCapRSound)
  from forward backward show ?thesis by simp
qed

(* one can pass a set of pairs to checkEdgRM,
  then tcap and tcap^-1 will be precomputed for all these pairs.
  Note, that the lookup might require more time than to actually compute the values, if root-symbol comparison suffices.
  Setting P = [] essentially disables the precomputation

  Moreover, if precomputeEdges is set to true, then even all edges will be precomputed.
  The latter might be useful for SCT where edge-tests are performed several times
*)
fun checkEdgRM :: "('f,'v)rule_map \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)rules \<Rightarrow> bool \<Rightarrow>  (('f,'v)rule \<times> ('f,'v)rule) \<Rightarrow> bool"
  where "checkEdgRM rm rrm P precomputeEdges = (if P = [] 
  then 
           (case rrm of None \<Rightarrow> (\<lambda> ((s,t),(u,v)). gcMatchI (capRM rm t) u) | _ \<Rightarrow> (\<lambda> ((s,t),(u,v)). gcMatchI (capRM rm t) u \<and> gcMatchI (capRM rrm u) t)) 
  else
     let forward = map (\<lambda> t. (t,capRM rm t)) (remdups (map snd P)) in
     case rrm of 
        None \<Rightarrow> if precomputeEdges then 
                 (let edges = (concat (map (\<lambda> (s,t). map (\<lambda> (u,v). ((t,u), gcMatchI (the (lookup t forward)) u)) P) P))
                   in (\<lambda> ((s,t),(u,v)). case lookup (t,u) edges of 
                            None \<Rightarrow> gcMatchI (capRM rm t) u
                         | Some b \<Rightarrow> b))
                else (\<lambda> ((s,t),(u,v)). gcMatchI (case lookup t forward of None \<Rightarrow> capRM rm t | Some ct \<Rightarrow> ct) u)
     | Some _ \<Rightarrow> let backward = map (\<lambda> u. (u,capRM rrm u)) (remdups (map fst P)) in
              if precomputeEdges then 
                 (let edges = (concat (map (\<lambda> (s,t). map (\<lambda> (u,v). ((t,u), gcMatchI (the (lookup t forward)) u \<and> gcMatchI (the (lookup u backward)) t)) P) P))
                   in (\<lambda> ((s,t),(u,v)). case lookup (t,u) edges of 
                            None \<Rightarrow> gcMatchI (capRM rm t) u \<and> gcMatchI (capRM rrm u) t
                         | Some b \<Rightarrow> b))
                else (\<lambda> ((s,t),(u,v)). gcMatchI (case lookup t forward of None \<Rightarrow> capRM rm t | Some ct \<Rightarrow> ct) u \<and> 
                                       gcMatchI (case lookup u backward of None \<Rightarrow> capRM rrm u | Some cu \<Rightarrow> cu) t ))"


lemma checkEdgRMSound[simp]: assumes seq: "(t \<cdot> \<sigma>, u \<cdot> \<delta>) \<in> (rstep (set R))^*" shows 
  "checkEdgRM (computeRuleMap R) (computeRuleMap (reverse_trs R)) P precomp ((s,t),(u,v))"
proof -
  from seq have forw: "gcMatch (capRM (computeRuleMap R) t) u" by (rule gcMatchCapRMSound)
  from seq have backw: "gcMatch (capRM (computeRuleMap (reverse_trs R)) u) t" by (rule gcMatchCapRRMSound)
  let ?forw = "map (\<lambda> t. (t, capRM (computeRuleMap R) t)) (remdups (map snd P))"
  let ?back = "map (\<lambda> u. (u, capRM (computeRuleMap (reverse_trs R)) u)) (remdups (map fst P))"
  let ?fedges = "(concat (map (\<lambda> (s,t). map (\<lambda> (u,v). ((t,u), gcMatch (the (lookup t ?forw)) u)) P) P))"
  let ?bedges = "(concat (map (\<lambda> (s,t). map (\<lambda> (u,v). ((t,u), gcMatch (the (lookup t ?forw)) u \<and> gcMatch (the (lookup u ?back)) t)) P) P))"
  let ?rrm = "computeRuleMap (reverse_trs R)"
  show ?thesis
  proof (cases P)
    case Nil
    with forw backw show ?thesis by (cases ?rrm, auto)
  next
    case (Cons pelem Ps)
    show ?thesis
    proof (cases ?rrm)
      case None
      show ?thesis 
      proof (cases "precomp")
	case False
	show ?thesis
	proof (cases "lookup t ?forw")
	  case None
	  with Cons False `?rrm = None` forw show ?thesis by auto
	next
	  case (Some ct)
	  with lookup_sound[where a = t and ps = ?forw and b = ct] have ct: "ct = capRM (computeRuleMap R) t" by auto
	  from Cons Some None False forw ct show ?thesis by (auto simp: Let_def ct)
	qed
      next
	case True
	show ?thesis
	proof (cases "lookup (t,u) ?fedges")
	  case None
	  with Cons True `?rrm = None` forw show ?thesis by (auto simp: Let_def)
	next
	  case (Some b)
	  hence "((t,u),b) \<in> set ?fedges" by (rule lookup_sound)
	  hence b: "b = gcMatchI (the (lookup t ?forw)) u" and t: "t \<in> snd ` set P" by (auto, force)
	  have lookup: "the (lookup t ?forw) = capRM (computeRuleMap R) t"
	  proof (cases "lookup t ?forw")
	    case (Some ct)
	    with lookup_sound[where a = t and ps = ?forw and b = ct] have ct: "ct = capRM (computeRuleMap R) t" by auto
	    with Cons Some show ?thesis by auto
	  next
	    case None
	    with lookup_complete[where a = t and ps = ?forw] t have False by auto
	    thus ?thesis ..
	  qed
	  with Cons True Some forw None b show ?thesis by (auto simp: Let_def)
	qed
      qed
    next
      case (Some rrm)
      show ?thesis 
      proof (cases "precomp")
	case False
	have idOne: "(case lookup t ?forw of None \<Rightarrow> capRM (computeRuleMap R) t | Some ct \<Rightarrow> ct) = (capRM (computeRuleMap R) t)"
	proof (cases "lookup t ?forw", simp)
	  case (Some ct)
	  with lookup_sound[where a = t and ps = ?forw and b = ct] have ct: "ct = capRM (computeRuleMap R) t" by auto
	  with Some show ?thesis by auto
	qed
	have idTwo: "(case lookup u ?back of None \<Rightarrow> capRM ?rrm u | Some ct \<Rightarrow> ct) = (capRM ?rrm u)"
	proof (cases "lookup u ?back", simp)
	  case (Some ct)
	  with lookup_sound[where a = u and ps = ?back and b = ct] have ct: "ct = capRM ?rrm u" by auto
	  with Some show ?thesis by auto
	qed
	from idOne idTwo False forw backw Some show ?thesis by (auto simp: Let_def Some, simp only: Some)
      next
	case True
	show ?thesis
	proof (cases "lookup (t,u) ?bedges")
	  case None
	  with True Some forw backw show ?thesis by (auto simp: Let_def)
	next
	  case (Some b)
	  hence "((t,u),b) \<in> set ?bedges" by (rule lookup_sound)
	  hence b: "b = (gcMatchI (the (lookup t ?forw)) u \<and> gcMatchI (the (lookup u ?back)) t)" and t: "t \<in> snd ` set P" and u: "u \<in> fst ` set P" by (auto, force+)
	  have lookup: "the (lookup t ?forw) = capRM (computeRuleMap R) t"
	  proof (cases "lookup t ?forw")
	    case (Some ct)
	    with lookup_sound[where a = t and ps = ?forw and b = ct] have ct: "ct = capRM (computeRuleMap R) t" by auto
	    with Some show ?thesis by auto
	  next
	    case None
	    with lookup_complete[where a = t and ps = ?forw] t have False by auto
	    thus ?thesis ..
	  qed
	  have lookup2: "the (lookup u ?back) = capRM ?rrm u"
	  proof (cases "lookup u ?back")
	    case (Some ct)
	    with lookup_sound[where a = u and ps = ?back and b = ct] have ct: "ct = capRM ?rrm u" by auto
	    with Some show ?thesis by auto
	  next
	    case None
	    with lookup_complete[where a = u and ps = ?back] u have False by auto
	    thus ?thesis ..
	  qed
	  from True `?rrm = Some rrm` Some forw backw b lookup lookup2 Cons show ?thesis by (auto simp: Let_def)
	qed
      qed
    qed
  qed
qed	  



declare checkEdg.simps[simp del]
declare checkEdgRM.simps[simp del]

notation Term.Hole ("\<box>")
no_notation HHole ("\<box>")
no_notation equiv_class ("[_]")


end
