(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory "Q-Restricted_Rewriting"
imports Trs
begin

abbreviation
  NF_terms :: "('f, 'v) term set \<Rightarrow> ('f, 'v) term set"
where
  "NF_terms Q \<equiv> NF (rstep (Id_on Q))"


lemma NF_map_vars_inj:
  assumes inj: "\<And> x. n (m x) = x" and NF: "t \<in> NF_terms Q"
  shows "(map_vars m t) \<in> NF_terms (map_vars m ` Q)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain u where "(map_vars m t, u) \<in> rstep (Id_on (map_vars m ` Q))" by blast
  then obtain ml mr C \<sigma> where in_mR: "(ml, mr) \<in> Id_on (map_vars m ` Q)"
    and mt: "map_vars m t = C\<langle>ml \<cdot> \<sigma>\<rangle>" by best
  let ?m = n
  from in_mR obtain l r where "(l, r) \<in> Id_on Q" and ml: "ml = map_vars m l" by auto
  have "t = map_vars ?m (map_vars m t)" by (simp add: map_vars_inj_compose[of n m, OF inj])
  also have "\<dots> = map_vars ?m (C\<langle>ml \<cdot> \<sigma>\<rangle>)" by (simp add: mt)
  also have "\<dots> = (map_vars_ctxt ?m C)\<langle>map_vars ?m (map_vars m l \<cdot> \<sigma>)\<rangle>"
    by (simp add: map_vars_ctxt_commute ml)
  also have "\<dots> = (map_vars_ctxt ?m C)\<langle>l \<cdot> (map_vars_subst_ran ?m (map_subst_domain m \<sigma>))\<rangle>"
    by (simp add: map_subst_domain map_vars_subst_ran)
  finally show False using NF and `(l, r) \<in> Id_on Q` by auto
qed


lemma notin_NF_terms: "t \<in> Q \<Longrightarrow> t \<notin> NF_terms Q"
using lhs_notin_NF_rstep[of t t "Id_on Q"] by (simp add: Id_on_iff)


subsection {* Q-Restricted Rewriting *}

inductive_set
  qrstep :: "('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) term ars"
  for Q and R
where
  id[intro]: "\<forall>u\<lhd>s. u \<in> NF_terms Q \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> (s, t) \<in> qrstep Q R" |
  ctxt[intro]: "(s, t) \<in> qrstep Q R \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> qrstep Q R" |
  subst[intro]:
    "\<forall>u\<lhd>(s \<cdot> \<sigma>). u \<in> NF_terms Q \<Longrightarrow> (s, t) \<in> qrstep Q R \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> qrstep Q R"

hide_fact (open)
  qrstep.id qrstep.ctxt qrstep.subst
  qrstepp.id qrstepp.ctxt qrstepp.subst

(* with the special case innermost rewriting *)
definition
  irstep :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term ars"
where
  "irstep R \<equiv> qrstep (lhss R) R"

(* and the special case standard rewriting *)
lemma rstep_qrstep_conv: "rstep R = qrstep {} R"
proof (intro equalityI subsetI2)
  fix s t assume "(s, t) \<in> rstep R" thus "(s, t) \<in> qrstep {} R"
  by (induct rule: rstep.induct) auto
next
  fix s t assume "(s, t) \<in> qrstep {} R" thus "(s, t) \<in> rstep R"
  by (induct) auto
qed

text {*
The inductive definition really corresponds to the intuitive definition of
Q-restricted rewriting.
*}
lemma qrstepE':
  assumes "(s, t) \<in> qrstep Q R"
  shows "\<exists>C \<sigma> l r. (\<forall>u\<lhd>l\<cdot>\<sigma>. u \<in> NF_terms Q) \<and> (l, r) \<in> R \<and> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle>"
using assms proof (induct rule: qrstep.induct)
  case (id s t)
  hence "\<forall>u\<lhd>s\<cdot>Var. u \<in> NF_terms Q" by auto
  moreover have "(s, t) \<in> R" by fact
  moreover have "s = \<box>\<langle>s \<cdot> Var\<rangle>" by simp
  moreover have "t = \<box>\<langle>t \<cdot> Var\<rangle>" by simp
  ultimately show ?case
    using exI[of _ \<box>, OF exI[of _ Var, OF exI[of _ s, OF exI[of _ t]]], of
    "\<lambda>\<sigma> C l r. (\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q) \<and> (l, r) \<in> R \<and> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle>"]
    by best
next
  case (ctxt s t C)
  then obtain D \<sigma> l r where "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q"
    and "(l, r) \<in> R" and s: "s = D\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = D\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  let ?C = "C \<circ>c D"
  have "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" by fact
  moreover have "(l, r) \<in> R" by fact
  moreover have "C\<langle>s\<rangle> = ?C\<langle>l \<cdot> \<sigma>\<rangle>" by (simp add: s)
  moreover have "C\<langle>t\<rangle> = ?C\<langle>r \<cdot> \<sigma>\<rangle>" by (simp add: t)
  ultimately show ?case
    using exI[of _ ?C, OF exI[of _ \<sigma>, OF exI[of _ l], OF exI[of _ r]], of
    "\<lambda>\<sigma> E l r. (\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q) \<and> (l, r) \<in> R \<and> C\<langle>s\<rangle> = E\<langle>l \<cdot> \<sigma>\<rangle> \<and> C\<langle>t\<rangle> = E\<langle>r \<cdot> \<sigma>\<rangle>"]
    by best
next
  case (subst s \<sigma> t)
  then obtain C \<tau> l r where "\<forall>u\<lhd>l \<cdot> \<tau>. u \<in> NF_terms Q"
    and "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<tau>\<rangle>" and t: "t = C\<langle>r \<cdot> \<tau>\<rangle>" by auto
  let ?\<sigma> = "\<tau> \<circ>s \<sigma>"
  let ?C = "C \<cdot>c \<sigma>"
  from s have "s \<unrhd> l \<cdot> \<tau>" by auto
  hence "s \<cdot> \<sigma> \<unrhd> l \<cdot> ?\<sigma>" by auto
  have "\<forall>u\<lhd>l \<cdot> ?\<sigma>. u \<in> NF_terms Q"
  proof (intro allI impI)
    fix u assume "l \<cdot> ?\<sigma> \<rhd> u"
    with `s \<cdot> \<sigma> \<unrhd> l \<cdot> ?\<sigma>` have "s \<cdot> \<sigma> \<rhd> u" by (rule supteq_supt_trans)
    with subst show "u \<in> NF_terms Q" by auto
  qed
  moreover have "(l, r) \<in> R" by fact
  moreover have "s \<cdot> \<sigma> = ?C\<langle>l \<cdot> ?\<sigma>\<rangle>" by (simp add: s)
  moreover have "t \<cdot> \<sigma> = ?C\<langle>r \<cdot> ?\<sigma>\<rangle>" by (simp add: t)
  ultimately show ?case
    using exI[of _ ?C, OF exI[of _ ?\<sigma>, OF exI[of _ l], OF exI[of _  r]], of
    "\<lambda>\<tau> E l r. (\<forall>u\<lhd>l \<cdot> \<tau>. u \<in> NF_terms Q) \<and> (l, r) \<in> R \<and> s \<cdot> \<sigma> = E\<langle>l \<cdot> \<tau>\<rangle> \<and> t \<cdot> \<sigma> = E\<langle>r \<cdot> \<tau>\<rangle>"]
    by best
qed

lemma qrstepE[elim]:
  assumes "(s, t) \<in> qrstep Q R"
    and "\<And>C \<sigma> l r. \<lbrakk>\<forall>u\<lhd>l\<cdot>\<sigma>. u \<in> NF_terms Q; (l, r) \<in> R; s = C\<langle>l \<cdot> \<sigma>\<rangle>; t = C\<langle>r \<cdot> \<sigma>\<rangle>\<rbrakk> \<Longrightarrow> P"
  shows "P"
using qrstepE'[of s t Q R] and assms by auto

lemma qrstepI[intro]:
  assumes "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q"
    and "(l, r) \<in> R"
    and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "(s, t) \<in> qrstep Q R"
proof -
  have "\<forall>u\<lhd>l. u \<in> NF_terms Q"
  proof (rule ccontr)
    assume "\<not> (\<forall>u\<lhd>l. u \<in> NF_terms Q)"
    then obtain u where "l \<rhd> u" and "u \<notin> NF_terms Q" by auto
    with NF_instance[of u \<sigma> "Id_on Q"] have "u \<cdot> \<sigma> \<notin> NF_terms Q" by auto
    moreover from `l \<rhd> u` have "l \<cdot> \<sigma> \<rhd> u \<cdot> \<sigma>" by (rule supt_stable)
    ultimately show False using assms by auto
  qed
  with `(l, r) \<in> R` have "(l, r) \<in> qrstep Q R" by auto
  with assms have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> qrstep Q R" by auto
  thus ?thesis unfolding s t by auto
qed

lemma qrstep_induct[case_names IH, induct set: qrstep]:
  assumes "(s, t) \<in> qrstep Q R"
    and IH: "\<And>C \<sigma> l r. \<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q \<Longrightarrow> (l, r) \<in> R \<Longrightarrow> P C\<langle>l \<cdot> \<sigma>\<rangle> C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "P s t"
proof -
  from `(s, t) \<in> qrstep Q R` obtain C \<sigma> l r where NF: "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q"
    and "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  from IH[OF NF `(l, r) \<in> R`] show ?thesis unfolding s t .
qed

lemma qrstep_rules_mono:
  assumes "R \<subseteq> R'" shows "qrstep Q R \<subseteq> qrstep Q R'"
proof
  fix s t assume "(s, t) \<in> qrstep Q R"
  thus "(s, t) \<in> qrstep Q R'" using assms by (induct rule: qrstep.induct) auto
qed

lemma NF_terms_anti_mono:
  "Q \<subseteq> Q' \<Longrightarrow> NF_terms Q' \<subseteq> NF_terms Q"
using rstep_mono[THEN NF_anti_mono, of "Id_on Q" "Id_on Q'"] by auto

lemma qrstep_NF_anti_mono:
  assumes "Q \<subseteq> Q'" shows "qrstep Q' R \<subseteq> qrstep Q R"
proof
  fix s t assume "(s, t) \<in> qrstep Q' R"
  thus "(s, t) \<in> qrstep Q R"
  proof (induct rule: qrstep.induct)
    case (id s t)
    from NF_terms_anti_mono[OF assms]
      have "\<forall>u\<lhd>s. u \<in> NF_terms Q" using id by auto
    with id show ?case by auto
  next
    case (ctxt s t C) thus ?case by auto
  next
    case (subst s \<sigma> t)
    from NF_terms_anti_mono[OF assms]
      have "\<forall>u\<lhd>s \<cdot> \<sigma>. u \<in> NF_terms Q" using subst by auto
    with subst show ?case by auto
  qed
qed

(* Lemma 2.4 in René's thesis *)
lemma qrstep_mono:
  assumes 1: "R \<subseteq> R'" and 2: "NF_terms Q \<subseteq> NF_terms Q'"
  shows "qrstep Q R \<subseteq> qrstep Q' R'"
proof
  fix s t assume "(s, t) \<in> qrstep Q R"
  then obtain C \<sigma> l r where "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" and "(l, r) \<in> R"
    and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  moreover with 2 have "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q'" by auto
  ultimately have "(s, t) \<in> qrstep Q' R" by auto
  with qrstep_rules_mono[OF 1] show "(s, t) \<in> qrstep Q' R'" ..
qed

(* Lemma 2.6 in René's thesis *)
lemma NF_terms_subset_criterion:
  "Q' \<inter> NF_terms Q = {} \<longleftrightarrow> NF_terms Q \<subseteq> NF_terms Q'" (is "?lhs = ?rhs")
proof
  assume "?rhs" thus "?lhs"
  proof (rule contrapos_pp)
    assume "\<not> ?lhs"
    then obtain q where "q \<in> Q'" and "q \<in> NF_terms Q" by auto
    from `q \<in> Q'` have "q \<notin> NF_terms Q'" by auto
    with `q \<in> NF_terms Q` show "\<not> NF_terms Q \<subseteq> NF_terms Q'" by blast
  qed
next
  assume "?lhs" thus "?rhs"
  proof (rule contrapos_pp)
    assume "\<not> ?rhs"
    then obtain t where "t \<in> NF_terms Q" and "t \<notin> NF_terms Q'" by auto
    then obtain s where "(t, s) \<in> rstep (Id_on Q')" by (auto simp: NF_def)
    then obtain C q \<sigma> where "q \<in> Q'" and "t = C\<langle>q \<cdot> \<sigma>\<rangle>" by auto
    have "q \<in> NF_terms Q"
    proof
      fix r show "(q, r) \<notin> rstep (Id_on Q)"
      proof
        assume "(q, r) \<in> rstep (Id_on Q)"
        hence "(q \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rstep (Id_on Q)" by auto
        hence "(t, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep (Id_on Q)" unfolding `t = C\<langle>q \<cdot> \<sigma>\<rangle>` by auto
        hence "t \<notin> NF_terms Q" by auto
        with `t \<in> NF_terms Q` show False by blast
      qed
    qed
    with `q \<in> Q'` show "Q' \<inter> NF_terms Q \<noteq> {}" by auto
  qed
qed

lemma qrstep_subset_rstep: "qrstep Q R \<subseteq> rstep R"
by (simp only: rstep_qrstep_conv, rule qrstep_NF_anti_mono, auto)

lemma NF_terms_lhss[simp]: "NF_terms (lhss R) = NF (rstep R)"
proof
  show "NF (rstep R) \<subseteq> NF_terms (lhss R)" by force
next
  show "NF_terms (lhss R) \<subseteq> NF (rstep R)"
  proof
    fix s assume NF: "s \<in> NF_terms (lhss R)"
    show "s \<in> NF (rstep R)"
    proof (rule ccontr)
      assume "s \<notin> NF (rstep R)"
      then obtain t where "(s, t) \<in> rstep R" by auto
      then obtain l r C \<sigma> where "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" by auto
      hence "(l, l) \<in> Id_on (lhss R)" by force
      hence "(s, s) \<in> rstep (Id_on (lhss R))" unfolding s by auto
      with NF show False by auto
    qed
  qed
qed

lemma NF_rstep_supt_args_conv:
  "(\<forall>u\<lhd>t. u \<in> NF (rstep R)) = (\<forall>u\<in>set (args t). u \<in> NF (rstep R))"
proof (cases t)
  case (Var x) show ?thesis unfolding Var by auto
next
  case (Fun f ts)
  show ?thesis (is "?lhs = ?rhs")
  proof
    assume "?lhs" thus "?rhs" by (auto simp: Fun)
  next
    assume "?rhs" show "?lhs"
    proof (intro allI impI)
      fix u assume "u \<lhd> t"
      from supt_Fun_imp_arg_supteq[OF this[unfolded Fun]]
        obtain t where "t \<in> set ts" and "t \<unrhd> u" by best
      from `?rhs`[unfolded Fun] and `t \<in> set ts` have "t \<in> NF (rstep R)" by simp
      from this and `t \<unrhd> u` show "u \<in> NF (rstep R)" by (rule NF_subterm)
    qed
  qed
qed

lemma qrstep_closed_ctxt:
  assumes "(s, t) \<in> qrstep Q R" shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> qrstep Q R"
using assms by auto

lemma ctxt_closed_qrstep: "ctxt_closed (qrstep Q R)"
unfolding ctxt_closed_def
proof (rule subsetI2)
  fix s t
  assume "(s, t) \<in> CC (qrstep Q R)"
  thus "(s, t) \<in> qrstep Q R" by induct auto
qed

lemma not_NF_rstep_minimal:
  assumes "s \<notin> NF (rstep R)"
  shows "\<exists>t\<unlhd>s. t \<notin> NF (rstep R) \<and> (\<forall>u\<lhd>t. u \<in> NF (rstep R))"
using assms proof (induct s rule: subterm_induct)
  case (subterm v)
  thus ?case
  proof (cases "\<forall>w\<lhd>v. w \<in> NF (rstep R)")
    case True with subterm(2) show ?thesis by auto
  next
    case False
    then obtain w where "v \<rhd> w" and "w \<notin> NF (rstep R)" by auto
    with subterm(1) obtain t where "w \<unrhd> t" and notNF: "t \<notin> NF (rstep R)"
      and NF: "\<forall>u\<lhd>t. u \<in> NF (rstep R)" by auto
    from `v \<rhd> w` and `w \<unrhd> t` have "v \<unrhd> t" using supt_supteq_trans[of v w t] by auto
    with notNF and NF show ?thesis by auto
  qed
qed

lemma rstep_imp_irstep:
  assumes "(s, t) \<in> rstep R" shows "\<exists>u. (s, u) \<in> irstep R"
proof -
  from assms have "s \<notin> NF (rstep R)" by auto
  from not_NF_rstep_minimal[OF this] obtain u where "s \<unrhd> u" and notNF: "u \<notin> NF (rstep R)"
    and NF: "\<forall>w\<lhd>u. w \<in> NF (rstep R)" by auto
  from notNF obtain v where "(u, v) \<in> rstep R" by auto
  then obtain l r C \<sigma> where "(l, r) \<in> R" and u: "u = C\<langle>l \<cdot> \<sigma>\<rangle>" and v: "v = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  from u have "u \<unrhd> l \<cdot> \<sigma>" by auto
  have "\<forall>w\<lhd>l \<cdot> \<sigma>. w \<in> NF (rstep R)"
  proof (intro allI impI)
    fix w assume "l \<cdot> \<sigma> \<rhd> w"
   with `u \<unrhd> l \<cdot> \<sigma>` have "u \<rhd> w" by (rule supteq_supt_trans)
   with NF show "w \<in> NF (rstep R)" by auto
  qed
  with `(l, r) \<in> R` and u and v have "(u, v) \<in> irstep R"
    unfolding irstep_def and NF_terms_lhss[symmetric, of R] by blast
  from `s \<unrhd> u` obtain C where "s = C\<langle>u\<rangle>" by auto
  from `(u, v) \<in> irstep R` have "(C\<langle>u\<rangle>, C\<langle>v\<rangle>) \<in> irstep R" unfolding irstep_def by auto
  thus ?thesis unfolding `s = C\<langle>u\<rangle>` by best
qed
  
lemma NF_irstep_NF_rstep: "NF (irstep R) = NF (rstep R)"
proof -
  from qrstep_NF_anti_mono[of "{}" "lhss R"]
    have "irstep R \<subseteq> rstep R" unfolding irstep_def rstep_qrstep_conv by blast
  from NF_anti_mono[OF this] have "NF (rstep R) \<subseteq> NF (irstep R)" .
  moreover have "NF (irstep R) \<subseteq> NF (rstep R)"
  proof
    fix s assume "s \<in> NF (irstep R)" show "s \<in> NF (rstep R)"
    proof (rule ccontr)
      assume "s \<notin> NF (rstep R)"
      then obtain t where "(s, t) \<in> rstep R" by auto
      from rstep_imp_irstep[OF this]
        obtain u where "(s, u) \<in> irstep R" by auto
      with `s \<in> NF (irstep R)` show False by auto
    qed
  qed
  ultimately show ?thesis by simp
qed

definition
  useful_rule :: "('f, 'v) term set \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool"
where
  "useful_rule Q lr \<equiv> \<forall>s\<lhd>fst lr. s \<in> NF_terms Q"

lemma useful_rule_empty: "useful_rule {} lr" 
unfolding useful_rule_def by auto

lemma only_useful_rules:
  assumes "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q"
  shows "useful_rule Q (l, r)"
unfolding useful_rule_def
proof (intro allI impI)
  fix s assume "fst (l, r) \<rhd> s"
  hence "l \<rhd> s" by simp
  hence "l \<cdot> \<sigma> \<rhd> s \<cdot> \<sigma>" by (rule supt_stable)
  with assms have "s \<cdot> \<sigma> \<in> NF_terms Q" by simp
  from NF_instance[OF this] show "s \<in> NF_terms Q" .
qed

definition
  useful_rules :: "('f, 'v) term set \<Rightarrow> ('f, 'v) trs  \<Rightarrow> ('f, 'v) trs"
where
  "useful_rules Q R \<equiv> {(l, r) | l r. (l, r) \<in> R \<and> useful_rule Q (l, r)}"

lemma useful_rules_empty:
  "useful_rules {} R = R"
unfolding useful_rules_def using useful_rule_empty by auto

lemma useful_rules_subset: "useful_rules Q R \<subseteq> R"
unfolding useful_rules_def useful_rule_def by auto

lemma qrstep_useful_rules: "qrstep Q (useful_rules Q R) = qrstep Q R"
proof -
  let ?U = "useful_rules Q R"
  show ?thesis
  proof
    show "qrstep Q ?U \<subseteq> qrstep Q R" by (rule qrstep_rules_mono[OF useful_rules_subset])
  next
    show "qrstep Q R \<subseteq> qrstep Q ?U"
    proof (rule subsetI2)
      fix s t
      assume "(s,t) \<in> qrstep Q R"
      then obtain l r C \<sigma> where "(l, r) \<in> R" and NF: "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q"
        and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
      with only_useful_rules[OF NF] have "(l, r) \<in> ?U" by (auto simp: useful_rules_def)
      with NF and `(l, r) \<in> R` and s and t show "(s, t) \<in> qrstep Q ?U" by auto
    qed
  qed
qed

text {* weakly well-formedness (all useful rules are well-formed). *}

definition
  wwf_rule :: "('f, 'v) term set \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool"
where
  "wwf_rule Q lr \<equiv>
    useful_rule Q lr \<longrightarrow> (\<not> is_var (fst lr) \<and> vars_term (snd lr) \<subseteq> vars_term (fst lr))"

definition
  wwf_qtrs :: "('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool"
where
  "wwf_qtrs Q R \<equiv> \<forall>(l, r)\<in>R. useful_rule Q (l, r) \<longrightarrow> (\<not> is_var l \<and> vars_term r \<subseteq> vars_term l)"

lemma wwf_qtrs_wwf_rules: "wwf_qtrs Q R = (\<forall>(l, r)\<in>R. wwf_rule Q (l, r))"
unfolding wwf_qtrs_def wwf_rule_def by auto

lemma wwf_qtrs_wf_trs: "wwf_qtrs Q R = wf_trs (useful_rules Q R)"
unfolding wwf_qtrs_def wf_trs_def useful_rules_def by force

lemma wwf_qtrs_empty: "wwf_qtrs {} R = wf_trs R"
unfolding wwf_qtrs_wf_trs useful_rules_empty by simp

lemma left_Var_imp_not_SN_qrstep:
  assumes "(Var x, r) \<in> R" shows "\<not> (SN_elt (qrstep Q R) t)"
proof -
  have steps: "\<forall>t. \<exists>s. (t, s) \<in> qrstep Q R"
  proof 
    fix t
    show "\<exists>s. (t, s) \<in> qrstep Q R"
    proof (induct t)
      case (Var y)
      let ?xt = "subst x (Var y)"
      let ?rxt = "r \<cdot> ?xt"
      have "\<forall>u\<lhd>Var x \<cdot> ?xt. u \<in> NF_terms Q" unfolding Var_subst by auto
      moreover from assms have "(Var x, r) \<in> qrstep Q R" by auto
      ultimately have "(Var x \<cdot> ?xt, ?rxt) \<in> qrstep Q R"
        using qrstep.subst[of "Var x" ?xt Q r] by auto
      thus ?case unfolding Var_subst by auto
    next
      case (Fun f ss)
      show ?case
      proof (cases ss)
        case Nil
        let ?xt = "subst x (Fun f ss)"
        let ?rxt = "r \<cdot> ?xt"
        have "\<forall>u\<lhd>Var x \<cdot> ?xt. u \<in> NF_terms Q" unfolding Nil Var_subst by auto
        moreover from assms have "(Var x, r) \<in> qrstep Q R" by auto
        ultimately have "(Var x \<cdot> ?xt, ?rxt) \<in> qrstep Q R"
          using qrstep.subst[of "Var x" ?xt Q r] by auto
        thus ?thesis unfolding Var_subst by auto
      next
        case (Cons s ts)
        with Fun obtain t where step: "(s, t) \<in> qrstep Q R" by auto
        let ?C = "More f [] \<box> ts"
        from Cons have id: "Fun f ss = ?C\<langle>s\<rangle>" by auto
        have "(?C\<langle>s\<rangle>, ?C\<langle>t\<rangle>) \<in> CC (qrstep Q R)"
           by (rule ctxt_closureI, rule step)
        with ctxt_closed_qrstep[of Q R]
        have "(Fun f ss, ?C\<langle>t\<rangle>) \<in> qrstep Q R"
          unfolding ctxt_closed_def id by auto
        thus ?thesis by auto
      qed
    qed
  qed
  show ?thesis 
  proof
    assume "SN_elt (qrstep Q R) t"
    from SN_elt_imp_normalizability[OF this]
      obtain s where norm: "(t, s) \<in> (qrstep Q R)^!" ..
    from steps obtain u where "(s, u) \<in> qrstep Q R" by auto
    with norm show False by auto
  qed
qed

lemma ctxt_compose_Suc: "(C ^ Suc i)\<langle>t\<rangle> = (C ^ i)\<langle>C\<langle>t\<rangle>\<rangle>"
using ctxt_power_compose_distr[of C i "Suc 0"] by simp

lemma SN_elt_imp_wwf_rule:
  assumes SN: "SN_elt (qrstep Q R) t"
    and t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" and lr: "(l, r) \<in> R" and NF: "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q"
  shows "wwf_rule Q (l, r)"
proof (cases "vars_term r \<subseteq> vars_term l")
  case True
  from left_Var_imp_not_SN_qrstep[of _ _ R Q t] lr SN have "\<not> is_var l" by (induct l) auto
  with True show ?thesis unfolding wwf_rule_def by auto
next
  case False
  then obtain x where "x \<in> vars_term r - vars_term l" by auto
  hence not_in_l: "x \<notin> vars_term l" by auto
  from only_useful_rules[OF NF] have "useful_rule Q (l,r)" .
  let ?\<delta> = "(\<lambda>y. if y = x then l \<cdot> \<sigma> else \<sigma> y)"
  have "l \<cdot> \<sigma> = l \<cdot> (\<sigma> |s (vars_term l))" by (rule coincidence_lemma)
  also have "\<dots> = l \<cdot> (?\<delta> |s (vars_term l))"
  proof (rule arg_cong[of _ _ "op \<cdot> l"])
    show "\<sigma> |s vars_term l = ?\<delta> |s vars_term l" 
    proof -
      {
        fix y
        have "(if y \<in> vars_term l then \<sigma> y else Var y) = (if y \<in> vars_term l then if y = x then l \<cdot> \<sigma> else \<sigma> y else Var y)" by (cases "y = x", auto simp: not_in_l)
      }
      note main = this
      show ?thesis unfolding subst_restrict_def by (rule ext, simp add: main)
    qed
  qed
  also have "\<dots> = l \<cdot> ?\<delta>" by (rule coincidence_lemma[symmetric])
  finally have lsld: "l \<cdot> \<sigma> = l \<cdot>?\<delta>" .
  from `x \<in> vars_term r - vars_term l` have in_r: "x \<in> vars_term r" by auto
  from supteq_Var[OF in_r] obtain C where "r = C\<langle>Var x\<rangle>" by auto
  hence "r \<cdot> ?\<delta> = (C \<cdot>c ?\<delta>)\<langle>l \<cdot> ?\<delta>\<rangle>" by (auto simp: lsld)
  then obtain C where rd:  "r \<cdot> ?\<delta> = C\<langle>l \<cdot> ?\<delta>\<rangle>" by auto
  obtain ls where ls: "l \<cdot> \<sigma> = ls" by auto
  obtain f where f: "\<And>i. f i = (C^i)\<langle>l \<cdot> ?\<delta>\<rangle>" by auto
  have steps: "\<forall>i. (f i, f (Suc i)) \<in> qrstep Q R"
  proof
    fix i
    show "(f i, f (Suc i)) \<in> qrstep Q R"
    proof
      show "\<forall>u\<lhd>l \<cdot> ?\<delta>. u \<in> NF_terms Q" unfolding lsld[symmetric] by fact
    next
      show "(l, r) \<in> R" by fact
    next
      show "f i = (C^i)\<langle>l \<cdot> ?\<delta>\<rangle>" by fact
    next
      show "f (Suc i) = (C ^ i)\<langle>r \<cdot> ?\<delta>\<rangle>" using f[of "Suc i", unfolded ctxt_compose_Suc rd[symmetric]] .
    qed
  qed
  have start: "f 0 = l \<cdot> \<sigma>" 
    by (simp add: f lsld[symmetric] ls one_ctxt_def)
  from start steps have nSN: "\<not> (SN_elt (qrstep Q R) (l \<cdot> \<sigma>))" unfolding SN_elt_def by auto
  from not_SN_subt_imp_not_SN[OF ctxt_closed_qrstep nSN ctxt_imp_supteq] SN[simplified t]
  have False ..
  thus ?thesis ..
qed

lemma SN_imp_wwf_qtrs:
  assumes "SN (qrstep Q R)" shows "wwf_qtrs Q R"
proof (rule ccontr)
  assume "\<not> wwf_qtrs Q R"
  then obtain l r where lr: "(l, r) \<in> R" and not_wwf: "\<not> wwf_rule Q (l, r)"
    unfolding wwf_qtrs_wwf_rules by auto
  hence useful: "useful_rule Q (l, r)" unfolding wwf_rule_def by auto
  from assms have SN: "SN_elt (qrstep Q R) l" unfolding SN_defs by auto
  from SN_elt_imp_wwf_rule[OF SN _ lr, of \<box> Var] and not_wwf
    have "\<not> (\<forall>u\<lhd>l. u \<in> NF_terms Q)" by auto
  then obtain u where "l \<rhd> u" and not_NF: "u \<notin> NF_terms Q" by auto
  from not_NF obtain v where "(u, v) \<in> rstep (Id_on Q)" by auto
  from `l \<rhd> u` obtain D where "D \<noteq> \<box>" and "l = D\<langle>u\<rangle>" by auto
  with useful[unfolded useful_rule_def] and `(u, v) \<in> rstep (Id_on Q)` show False by auto
qed

lemma left_Var_useful: "useful_rule Q (Var x,r)"
unfolding useful_rule_def using Var_supt[of x] by auto

lemma wwf_qtrs_imp_left_fun:
  assumes "wwf_qtrs Q R" and "(l, r) \<in> R" shows "\<exists>f ls. l = Fun f ls"
using assms unfolding wwf_qtrs_def
using left_Var_useful[of Q _ r] by (cases l) auto

definition
  rqrstep :: "('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) term ars"
where
  "rqrstep Q R =
    {(s, t). \<exists>l r \<sigma>. (\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q) \<and> (l, r) \<in> R \<and> s = l \<cdot> \<sigma> \<and> t = r \<cdot> \<sigma>}"

definition
  nrqrstep :: "('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) term ars"
where
  "nrqrstep Q R =
    {(s, t). \<exists>l r C \<sigma>.
    (\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q) \<and> (l, r) \<in> R \<and> C \<noteq> \<box> \<and> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle>}"

lemma rqrstepI[intro]:
  assumes "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" and "(l, r) \<in> R"
    and "s = l \<cdot> \<sigma>" and "t = r \<cdot> \<sigma>"
  shows "(s, t) \<in> rqrstep Q R"
using assms unfolding rqrstep_def by auto

lemma nrqrstepI[intro]:
  assumes "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" and "(l, r) \<in> R"
    and "C \<noteq> \<box>" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "(s, t) \<in> nrqrstep Q R"
using assms unfolding nrqrstep_def by auto

lemma rqrstepE[elim]:
  assumes "(s, t) \<in> rqrstep Q R"
    and "\<And>l r \<sigma>. \<lbrakk>\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q; (l, r) \<in> R; s = l \<cdot> \<sigma>; t = r \<cdot> \<sigma>\<rbrakk> \<Longrightarrow> P"
  shows "P"
using assms unfolding rqrstep_def by auto

lemma nrqrstepE[elim]:
  assumes "(s, t) \<in> nrqrstep Q R"
    and "\<And>l r C \<sigma>.
    \<lbrakk>\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q; (l, r) \<in> R; C \<noteq> \<box>; s = C\<langle>l \<cdot> \<sigma>\<rangle>; t = C\<langle>r \<cdot> \<sigma>\<rangle>\<rbrakk> \<Longrightarrow> P"
  shows "P"
using assms unfolding nrqrstep_def by auto

lemma qrstep_iff_rqrstep_or_nrqrstep:
  "qrstep Q R = rqrstep Q R \<union> nrqrstep Q R" (is "?step = ?root \<union> ?nonroot")
proof
  show "?step \<subseteq> ?root \<union> ?nonroot"
  proof
    fix s t assume "(s, t) \<in> ?step"
    then obtain C l r \<sigma> where NF: "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" and
       in_R: "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
    thus "(s, t) \<in> ?root \<union> ?nonroot" by (cases "C = \<box>") auto
  qed
next
  show "?root \<union> ?nonroot \<subseteq> ?step"
  proof (intro subsetI2, elim UnE)
    fix s t assume "(s, t) \<in> rqrstep Q R"
    then obtain l r \<sigma> where NF: "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" and
      in_R: "(l, r) \<in> R" and s: "s = \<box>\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = \<box>\<langle>r \<cdot> \<sigma>\<rangle>"
      unfolding ctxt_apply.simps by auto
    thus "(s, t) \<in> qrstep Q R" using qrstepI[of l \<sigma> Q r R s \<box> t] by simp
  next
    fix s t assume "(s, t) \<in> nrqrstep Q R" thus "(s, t) \<in> qrstep Q R" by auto
  qed
qed

lemma nrqrstep_preserves_root:
  fixes R :: "('f, 'v) trs"
  assumes "(Fun f ss, t) \<in> nrqrstep Q R" 
  shows "\<exists>ts. t = Fun f ts"
proof -
  from assms
    obtain C l r and \<sigma> :: "('f, 'v) subst"
    where t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and "C \<noteq> \<box>" and "Fun f ss = C\<langle>l \<cdot> \<sigma>\<rangle>" by auto
  then obtain ss1 D ss2 where "C = More f ss1 D ss2" by (cases C) auto
  with t have "t = Fun f (ss1 @ (D\<langle>r \<cdot> \<sigma>\<rangle> # ss2))" by auto
  thus ?thesis ..
qed

lemma nrqrsteps_preserve_root:
 assumes "(Fun f ss,t) \<in> (nrqrstep Q R)^*"
 shows "\<exists>ts. t = (Fun f ts)"
using assms by induct (auto simp: nrqrstep_preserves_root)

lemma nrqrstep_num_args:
  assumes "(s, t) \<in> nrqrstep Q R" shows "num_args s = num_args t"
proof -
  from assms obtain l r C \<sigma> where "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q"
    and "(l, r) \<in> R" and "C \<noteq> \<box>" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  show ?thesis unfolding s t by (cases C) (auto simp: `C \<noteq> \<box>`)
qed

lemma nrqrsteps_num_args:
  assumes "(s, t) \<in> (nrqrstep Q R)^*" shows "num_args s = num_args t"
using assms by (induct rule: rtrancl.induct) (auto simp: nrqrstep_num_args)


lemma nrqrstep_imp_arg_qrstep:
  assumes "(s, t) \<in> nrqrstep Q R"
  shows "(args s ! i, args t ! i) \<in> (qrstep Q R)^="
proof -
  from assms obtain l r C \<sigma> where NF: "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q"
    and in_R: "(l, r) \<in> R" and ne: "C \<noteq> \<box>" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  from ne obtain f ss1 D ss2 where C: "C = More f ss1 D ss2" by (cases C) auto
  let ?i = "length ss1"
  have "i < ?i \<or> i = ?i \<or> i > ?i" by auto
  thus ?thesis
  proof (elim disjE)
    assume "i < ?i"
    with append_Cons_nth_left[OF this] show ?thesis by (simp add: s t C)
  next
    assume "i = ?i"
    with append_Cons_nth_middle[OF this] show ?thesis using NF in_R s t by (auto simp: C)
  next
    assume "i > ?i"
    with append_Cons_nth_right[OF this] show ?thesis by (simp add: s t C)
  qed
qed

lemma nrqrsteps_imp_arg_qrsteps:
  assumes "(s, t) \<in> (nrqrstep Q R)^*" shows "(args s ! i, args t ! i) \<in> (qrstep Q R)^*"
using assms proof (induct rule: rtrancl_induct)
  case (step u v)
  with nrqrstep_imp_arg_qrstep[of u v Q R i] show ?case by auto
qed simp

lemma qrsteps_rqrstep_cases:
  assumes "(Fun f ss, t) \<in> (qrstep Q R)^*"
  shows "(\<exists>ts.
    length ts = length ss \<and>
    t = Fun f ts \<and>
    (\<forall>i<length ss. (ss ! i, ts ! i) \<in> (qrstep Q R)^*))
    \<or> (Fun f ss, t) \<in> (qrstep Q R)^* O (rqrstep Q R) O (qrstep Q R)^*"
using assms proof (induct)
  case base thus ?case by auto
next
  case (step u t)
  hence "(\<exists>ts. length ts = length ss \<and> u = Fun f ts
    \<and> (\<forall>i<length ss. (ss!i, ts!i) \<in> (qrstep Q R)^*))
    \<or> (Fun f ss, u) \<in> (qrstep Q R)^* O rqrstep Q R O (qrstep Q R)^*" by simp
  thus ?case
  proof
    assume "(Fun f ss, u) \<in> (qrstep Q R)^* O rqrstep Q R O (qrstep Q R)^*"
    then obtain v where "(Fun f ss, v) \<in> (qrstep Q R)^* O rqrstep Q R"
      and "(v, u) \<in> (qrstep Q R)^*" by auto
    with `(u, t) \<in> qrstep Q R` have "(v, t) \<in> (qrstep Q R)^*" by auto
    with `(Fun f ss, v) \<in> (qrstep Q R)^* O rqrstep Q R` show ?thesis by auto
  next
    assume "\<exists>ts. length ts = length ss \<and> u = Fun f ts
      \<and> (\<forall>i<length ss. (ss!i, ts!i) \<in> (qrstep Q R)^*)"
    then obtain ts where len: "length ts = length ss" and u: "u = Fun f ts"
      and steps: "\<forall>i<length ss. (ss!i, ts!i) \<in> (qrstep Q R)^*" by auto
    from `(u, t) \<in> qrstep Q R` have "(u, t) \<in> rqrstep Q R \<or> (u, t) \<in> nrqrstep Q R"
      unfolding qrstep_iff_rqrstep_or_nrqrstep by simp
    thus ?thesis
    proof
      assume "(u, t) \<in> rqrstep Q R"
      with `(Fun f ss, u) \<in> (qrstep Q R)^*` have "(Fun f ss, t) \<in> (qrstep Q R)^* O rqrstep Q R"
        by auto
      thus ?thesis by auto
    next
      assume "(u, t) \<in> nrqrstep Q R"
      from nrqrstep_preserves_root[OF this[unfolded u]]
        obtain us where t: "t = Fun f us" by auto
      from nrqrstep_num_args[OF `(u, t) \<in> nrqrstep Q R`]
        have "num_args u = num_args t" by simp
      hence "length ts = length us" unfolding u t by simp
      with len have len': "length ss = length ts" by simp
      from nrqrstep_imp_arg_qrstep[OF `(u, t) \<in> nrqrstep Q R`]
        have "\<forall>i<length ts. (ts!i, args t ! i) \<in> (qrstep Q R)^=" unfolding u by auto
      hence "\<forall>i<length ts. (ts!i, us!i) \<in> (qrstep Q R)^=" unfolding t by simp
      hence next_steps: "\<forall>i<length ts. (ts!i, us!i) \<in> (qrstep Q R)^*" by auto
      have "length us = length ss" using `length ts = length us` and len by simp
      moreover have "t = Fun f us" by fact
      moreover have "\<forall>i<length ss. (ss!i, us!i) \<in> (qrstep Q R)^*"
      proof (intro allI impI)
        fix i assume "i < length ss"
        with steps have "(ss ! i, ts ! i) \<in> (qrstep Q R)^*" by simp
        moreover from `i < length ss` and next_steps have "(ts!i, us!i) \<in> (qrstep Q R)^*"
          unfolding len' by simp
        ultimately show "(ss ! i, us ! i) \<in> (qrstep Q R)^*" by auto
      qed
      ultimately show ?thesis by auto
    qed
  qed
qed

end
