(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Term_Order_Impl
imports
  Term_Order
  Show
begin

locale redpair_impl = redpair + 
  fixes checkS :: "('f::show, 'v::show) rule \<Rightarrow> shows check"
    and checkNS :: "('f, 'v) rule \<Rightarrow> shows check"
  assumes checkSSound: "isOK (checkS rule) \<Longrightarrow> rule \<in> S" 
    and checkNSSound: "isOK (checkNS rule)  \<Longrightarrow> rule \<in> NS"

locale mono_redpair_impl = redpair_impl + 
  fixes checkMono :: "shows check"
  assumes checkMonoSound: "isOK(checkMono) \<Longrightarrow> one_step_ctxt_closed S"

locale ce_redpair_impl = redpair_impl + ce_redpair

locale ce_af_redpair_impl = redpair_impl + ce_af_redpair

sublocale ce_af_redpair_impl \<subseteq> ce_redpair_impl ..

locale ce_mono_redpair_impl = mono_ce_redpair + mono_redpair_impl

locale mono_ce_af_redpair_impl = ce_mono_redpair_impl + ce_af_redpair_impl


(* valid, check_strict, check_non_strict, check_mono, description *)
type_synonym ('f,'v)redpair_impl = "(
    shows check \<times>
    (('f :: show, 'v :: show)rule \<Rightarrow> shows check) \<times>
    (('f, 'v)rule \<Rightarrow> shows check) \<times>
    'f af \<times>
    shows check \<times>
    shows)"

fun rp_valid :: "('f,'v)redpair_impl \<Rightarrow> shows check"
where "rp_valid (v,s,n,a,m,d) = v"

fun rp_strict :: "('f,'v)redpair_impl \<Rightarrow> (('f, 'v)rule \<Rightarrow> shows check)"
where "rp_strict (v,s,n,a,m,d) = s"

fun rp_non_strict :: "('f,'v)redpair_impl \<Rightarrow> (('f, 'v)rule \<Rightarrow> shows check)"
where "rp_non_strict (v,s,n,a,m,d) = n"

fun rp_af :: "('f,'v)redpair_impl \<Rightarrow> 'f af"
where "rp_af (v,s,n,a,m,d) = a"

fun rp_mono :: "('f,'v)redpair_impl \<Rightarrow> shows check"
where "rp_mono (v,s,n,a,m,d) = m"

fun rp_desc :: "('f,'v)redpair_impl \<Rightarrow> shows"
where "rp_desc (v,s,n,a,m,d) = d"


locale generic_redpair_impl = 
  fixes generate_redpair :: "'a \<Rightarrow> ('f :: show,'v :: show)redpair_impl" 
  assumes generate_redpair: "isOK(rp_valid (generate_redpair rp)) 
  \<Longrightarrow> isOK (check_allm (rp_strict (generate_redpair rp)) s_list)
  \<Longrightarrow> isOK (check_allm (rp_non_strict (generate_redpair rp)) ns_list)
  \<Longrightarrow> \<exists> S NS. 
  ce_af_redpair S NS (rp_af (generate_redpair rp)) \<and> 
  set s_list \<subseteq> S \<and> 
  set ns_list \<subseteq> NS \<and> 
  (isOK (rp_mono (generate_redpair rp)) \<longrightarrow> mono_ce_af_redpair S NS (rp_af (generate_redpair rp)) \<and> one_step_ctxt_closed S)"

end